# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Temporary backend to be used to apply T2 and T1 noise.
"""
from typing import List

from qiskit import QuantumCircuit
from qiskit.circuit import Delay

from qiskit_aer import AerSimulator
from qiskit_aer.jobs.aerjob import AerJob
from qiskit_aer.noise.passes import RelaxationNoisePass

from qiskit_experiments.framework import BackendData


class NoisyDelayAerBackend(AerSimulator):
    """Backend for T1 and T2Ramsey experiments."""

    def __init__(
        self,
        t1: List[float] = None,
        t2: List[float] = None,
        dt: float = None,
        backend=None,
        **backend_options,
    ):
        """configure backend noise"""

        super().__init__(**backend_options)
        self._t2 = t2 or [1e-4]
        self._t1 = t1 or [2e-4]
        if backend:
            dt = BackendData(backend).dt
        if dt is not None:
            self._dt_unit = True
            self._dt_factor = dt
        else:
            self._dt_unit = False
            self._dt_factor = dt or 1e-9

        self._op_types = [Delay]

    # pylint: disable=arguments-differ
    def run(self, run_input: List[QuantumCircuit], **run_options) -> AerJob:
        """
        Add noise pass to all circuits and then run the circuits.

        Args:
            run_input: List of circuit to run.
            run_options (kwargs): additional run time backend options.

        Returns:
            AerJob: A job that contains the simulated data.

        """
        relax_pass = RelaxationNoisePass(
            self._t1, self._t2, dt=self._dt_factor, op_types=self._op_types
        )

        noisy_circuits = []
        for circ in run_input:
            noisy_circuits.append(relax_pass(circ))

        return super().run(noisy_circuits, **run_options)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Tphi Experiment class.
"""

from typing import List, Optional, Union, Sequence
import numpy as np

from qiskit import QiskitError
from qiskit.providers import Backend
from qiskit_experiments.framework.composite.batch_experiment import BatchExperiment
from qiskit_experiments.library.characterization import (
    T1,
    T2Ramsey,
    T2Hahn,
    TphiAnalysis,
)


class Tphi(BatchExperiment):
    r"""An experiment to measure the qubit dephasing rate in the :math:`x - y` plane.

    # section: overview

    :math:`T_\varphi`, or :math:`1/\Gamma_\varphi`, is the pure dephasing time in
    the :math:`x - y` plane of the Bloch sphere. We compute :math:`\Gamma_\varphi`
    by computing :math:`\Gamma_2`, the transverse relaxation rate, and subtracting
    :math:`\Gamma_1`, the longitudinal relaxation rate. It follows that

    :math:`1/T_\varphi = 1/T_2 - 1/2T_1`.

    The transverse relaxation rate can be estimated by either :math:`T_2` or
    :math:`T_2^*` experiments. In superconducting qubits, :math:`T_2^*` tends to be
    significantly smaller than :math:`T_1`, so :math:`T_2` is usually used.

    .. note::
        In 0.5.0, this experiment changed from using :math:`T_2^*` as the default
        to :math:`T_2`.

    # section: analysis_ref
        :class:`.TPhiAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            # Temporary workaround for missing support in Qiskit and qiskit-ibm-runtime
            from qiskit_experiments.test.patching import patch_sampler_test_support
            patch_sampler_test_support()

            # backend
            from qiskit_ibm_runtime.fake_provider import FakeManilaV2
            from qiskit_aer import AerSimulator
            from qiskit_aer.noise import NoiseModel

            noise_model = NoiseModel.from_backend(FakeManilaV2(),
                                                  thermal_relaxation=True,
                                                  gate_error=False,
                                                  readout_error=False,
                                                 )

            backend = AerSimulator.from_backend(FakeManilaV2(), noise_model=noise_model)

        .. jupyter-execute::

            import numpy as np
            import qiskit
            from qiskit_experiments.library.characterization import Tphi

            delays_t1 = np.arange(1e-6, 300e-6, 10e-6)
            delays_t2 = np.arange(1e-6, 50e-6, 2e-6)

            exp = Tphi(physical_qubits=(0, ),
                        delays_t1=delays_t1,
                        delays_t2=delays_t2,
                        backend=backend
                        )

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            display(exp_data.figure(1))
            exp_data.analysis_results(dataframe=True)

    # section: reference
        .. ref_arxiv:: 1 1904.06560

    # section: manual
        :doc:`/manuals/characterization/tphi`

    # section: see_also
        * :py:class:`qiskit_experiments.library.characterization.T1`
        * :py:class:`qiskit_experiments.library.characterization.T2Ramsey`
        * :py:class:`qiskit_experiments.library.characterization.T2Hahn`

    """

    def __init__(
        self,
        physical_qubits: Sequence[int],
        delays_t1: List[Union[List[float], np.array]],
        delays_t2: List[Union[List[float], np.array]],
        t2type: str = "hahn",
        osc_freq: float = 0.0,
        num_echoes: int = 1,
        backend: Optional[Backend] = None,
    ):
        """Initialize the experiment object.

        Args:
            physical_qubits: A single-element sequence containing the qubit under test.
            t2type: What type of T2/T2* experiment to use. Can be either "ramsey" for
                :class:`.T2Ramsey` to be used, or "hahn" for :class:`.T2Hahn`. Defaults
                to "hahn".
            delays_t1: Delay times of the T1 experiment.
            delays_t2: Delay times of the T2 experiment.
            osc_freq: The oscillation frequency induced for T2Ramsey. Only used when
                ``t2type`` is set to "ramsey".
            num_echoes: The number of echoes to perform for T2Hahn. Only used when
                ``t2type`` is set to "hahn".
            backend: Optional, the backend on which to run the experiment.

        Raises:
            QiskitError: If an invalid ``t2type`` is provided.
        """

        exp_t1 = T1(physical_qubits=physical_qubits, delays=delays_t1, backend=backend)

        exp_options = {"delays_t1": delays_t1, "delays_t2": delays_t2}

        if t2type == "ramsey":
            exp_t2 = T2Ramsey(
                physical_qubits=physical_qubits,
                delays=delays_t2,
                backend=backend,
                osc_freq=osc_freq,
            )
            exp_options["osc_freq"] = osc_freq
        elif t2type == "hahn":
            exp_t2 = T2Hahn(
                physical_qubits=physical_qubits,
                delays=delays_t2,
                backend=backend,
                num_echoes=num_echoes,
            )
            exp_options["num_echoes"] = num_echoes
        else:
            raise QiskitError(f"Invalid T2 experiment type {t2type} specified.")

        analysis = TphiAnalysis([exp_t1.analysis, exp_t2.analysis])

        # Create batch experiment
        super().__init__(
            [exp_t1, exp_t2],
            flatten_results=True,
            backend=backend,
            analysis=analysis,
        )
        self.set_experiment_options(**exp_options)

    def set_experiment_options(self, **fields):
        """Set the experiment options.
        Args:
            fields: The fields defining the options.

        Raises:
             QiskitError: Invalid input option.
        """
        # propagate options to the sub-experiments.
        for key in fields:
            if key == "delays_t1":
                self.component_experiment(0).set_experiment_options(delays=fields["delays_t1"])
            elif key == "delays_t2":
                self.component_experiment(1).set_experiment_options(delays=fields["delays_t2"])
            elif key == "osc_freq" and isinstance(self.component_experiment(1), T2Ramsey):
                self.component_experiment(1).set_experiment_options(osc_freq=fields["osc_freq"])
            elif key == "num_echoes" and isinstance(self.component_experiment(1), T2Hahn):
                self.component_experiment(1).set_experiment_options(num_echoes=fields["num_echoes"])
            else:
                raise QiskitError(f"Tphi experiment does not support option {key}.")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
===============================================================================================
Driven Frequency Tuning (:mod:`qiskit_experiments.library.driven_freq_tuning`)
===============================================================================================

.. currentmodule:: qiskit_experiments.library.driven_freq_tuning

Experiments
===========
.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/experiment.rst

    StarkRamseyXY
    StarkRamseyXYAmpScan
    StarkP1Spectroscopy


Analysis
========

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/analysis.rst

    StarkRamseyXYAmpScanAnalysis
    StarkP1SpectAnalysis


Stark Coefficient
=================

.. autosummary::
    :toctree: ../stubs/

    StarkCoefficients
    retrieve_coefficients_from_backend
    retrieve_coefficients_from_service
"""

from .ramsey_amp_scan_analysis import StarkRamseyXYAmpScanAnalysis
from .p1_spect_analysis import StarkP1SpectAnalysis
from .ramsey import StarkRamseyXY
from .ramsey_amp_scan import StarkRamseyXYAmpScan
from .p1_spect import StarkP1Spectroscopy

from .coefficients import (
    StarkCoefficients,
    retrieve_coefficients_from_backend,
    retrieve_coefficients_from_service,
)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Rabi amplitude experiment."""

from typing import Iterable, List, Optional, Sequence, Tuple
import numpy as np

from qiskit import QuantumCircuit
from qiskit.circuit import Gate, Parameter
from qiskit.qobj.utils import MeasLevel
from qiskit.providers import Backend
from qiskit.pulse import ScheduleBlock
from qiskit.exceptions import QiskitError
from qiskit.utils.deprecation import deprecate_func

from qiskit_experiments.framework import BaseExperiment, Options
from qiskit_experiments.framework.restless_mixin import RestlessMixin
from qiskit_experiments.curve_analysis import ParameterRepr, OscillationAnalysis


class Rabi(BaseExperiment, RestlessMixin):
    r"""An experiment that scans a pulse amplitude to calibrate rotations on the :math:`|0\rangle`
    <-> :math:`|1\rangle` transition.

    # section: overview

        The circuits have a custom rabi gate with the pulse schedule attached to it
        through the calibrations. The circuits are of the form:

        .. parsed-literal::

                       ┌───────────┐ ░ ┌─┐
                  q_0: ┤ Rabi(amp) ├─░─┤M├
                       └───────────┘ ░ └╥┘
            measure: 1/═════════════════╩═
                                        0

        The user provides his own schedule for the Rabi at initialization which must have one
        free parameter, i.e. the amplitude to scan and a drive channel which matches the qubit.

    # section: manual
        :ref:`Rabi Calibration`

        See also the `Qiskit Textbook
        <https://github.com/Qiskit/textbook/blob/main/notebooks/quantum-hardware-pulses/calibrating-qubits-pulse.ipynb>`_
        for the pulse level programming of a Rabi experiment.

    # section: analysis_ref
        :class:`~qiskit_experiments.curve_analysis.OscillationAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            import warnings

            warnings.filterwarnings(
                "ignore",
                message=".*Due to the deprecation of Qiskit Pulse.*",
                category=DeprecationWarning,
            )

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9,-0.25e9, 1e9, 0.8e9, 1e4, noise=True, seed=199)

        .. jupyter-execute::

            import numpy as np
            from qiskit import pulse
            from qiskit.circuit import Parameter
            from qiskit_experiments.library import Rabi

            with pulse.build() as build_sched:
                pulse.play(pulse.Gaussian(160, Parameter("amp"), 40), pulse.DriveChannel(0))

            exp = Rabi(physical_qubits=(0,),
                       schedule=build_sched,
                       amplitudes=np.linspace(-0.1, 0.1, 21),
                       backend=backend,)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)
    """

    __gate_name__ = "Rabi"
    __outcome__ = "rabi_rate"

    @classmethod
    def _default_run_options(cls) -> Options:
        """Default option values for the experiment :meth:`run` method."""
        options = super()._default_run_options()

        options.meas_level = MeasLevel.KERNELED
        options.meas_return = "single"

        return options

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default values for the pulse if no schedule is given.

        Experiment Options:
            amplitudes (iterable): The list of amplitude values to scan.
            schedule (ScheduleBlock): The schedule for the Rabi pulse. This schedule must have
                exactly one free parameter. The drive channel should match the qubit.

        """
        options = super()._default_experiment_options()

        options.amplitudes = np.linspace(-0.95, 0.95, 51)
        options.schedule = None

        return options

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, experiments involving pulse "
            "gate calibrations like this one have been deprecated."
        ),
    )
    def __init__(
        self,
        physical_qubits: Sequence[int],
        schedule: ScheduleBlock,
        amplitudes: Optional[Iterable[float]] = None,
        backend: Optional[Backend] = None,
    ):
        """Initialize a Rabi experiment on the given qubit.

        Args:
            physical_qubits: List with the qubit on which to run the Rabi experiment.
            schedule: The schedule that will be used in the Rabi experiment. This schedule
                should have one free parameter namely the amplitude.
            amplitudes: The pulse amplitudes that one wishes to scan. If this variable is not
                specified it will default to :code:`np.linspace(-0.95, 0.95, 51)`.
            backend: Optional, the backend to run the experiment on.
        """
        super().__init__(physical_qubits, analysis=OscillationAnalysis(), backend=backend)

        self.analysis.set_options(
            result_parameters=[ParameterRepr("freq", self.__outcome__)],
            normalization=True,
        )
        self.analysis.plotter.set_figure_options(
            xlabel="Amplitude",
            ylabel="Signal (arb. units)",
        )

        if amplitudes is not None:
            self.experiment_options.amplitudes = amplitudes

        self.experiment_options.schedule = schedule

    def _pre_circuit(self) -> QuantumCircuit:
        """A circuit with operations to perform before the Rabi."""
        return QuantumCircuit(1)

    def _template_circuit(self) -> Tuple[QuantumCircuit, Parameter]:
        """Return the template quantum circuit."""
        sched = self.experiment_options.schedule
        param = next(iter(sched.parameters))

        if len(sched.parameters) != 1:
            raise QiskitError(
                f"Schedule {sched} for {self.__class__.__name__} experiment must have "
                f"exactly one free parameter, found {sched.parameters} parameters."
            )

        gate = Gate(name=self.__gate_name__, num_qubits=1, params=[param])

        circuit = self._pre_circuit()
        circuit.append(gate, (0,))
        circuit.measure_active()
        circuit.add_calibration(gate, self._physical_qubits, sched, params=[param])

        return circuit, param

    def circuits(self) -> List[QuantumCircuit]:
        """Create the circuits for the Rabi experiment.

        Returns:
            A list of circuits with a rabi gate with an attached schedule. Each schedule
            will have a different value of the scanned amplitude.
        """

        # Create template circuit
        circuit, param = self._template_circuit()

        # Create the circuits to run
        circs = []
        for amp in self.experiment_options.amplitudes:
            # casting is needed because for amplitude '0', np.round method return datatype of int32
            # which isn't serializable in the metadata.
            amp = float(np.round(amp, decimals=6))
            assigned_circ = circuit.assign_parameters({param: amp}, inplace=False)
            assigned_circ.metadata = {"xval": amp}

            circs.append(assigned_circ)

        return circs

    def _metadata(self):
        metadata = super()._metadata()
        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)
        return metadata


class EFRabi(Rabi):
    r"""An experiment that scans the amplitude of a pulse inducing rotations on the
    :math:`|1\rangle` <-> :math:`|2\rangle` transition.

    # section: overview

        This experiment is a subclass of the :class:`Rabi` experiment but takes place between
        the first and second excited state. An initial X gate populates the first excited state.
        The Rabi pulse is applied on the :math:`|1\rangle` <-> :math:`|2\rangle` transition
        (sometimes also labeled the e <-> f transition). The necessary frequency shift (typically
        the qubit anharmonicity) is given through the pulse schedule given at initialization. The
        schedule is then also stored in the experiment options. The circuits are of the form:

        .. parsed-literal::

                       ┌───┐┌───────────┐ ░ ┌─┐
                  q_0: ┤ X ├┤ Rabi(amp) ├─░─┤M├
                       └───┘└───────────┘ ░ └╥┘
            measure: 1/══════════════════════╩═
                                             0
    # section: example
        .. jupyter-execute::
            :hide-code:

            import warnings

            warnings.filterwarnings(
                "ignore",
                message=".*Due to the deprecation of Qiskit Pulse.*",
                category=DeprecationWarning,
            )

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9, -0.25e9, 1e9, 0.8e9, 1e4, noise=True, seed=198)

        .. jupyter-execute::

            import numpy as np
            from qiskit import pulse
            from qiskit.circuit import Parameter
            from qiskit_experiments.library import EFRabi

            with pulse.build() as build_sched:
                pulse.play(pulse.Gaussian(160, Parameter("amp"), sigma=40), pulse.DriveChannel(0))

            exp = EFRabi(physical_qubits=(0,),
                         backend=backend,
                         schedule=build_sched,
                         amplitudes=np.linspace(-0.1, 0.1, 21),)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)
    """

    __outcome__ = "rabi_rate_12"

    def _pre_circuit(self) -> QuantumCircuit:
        """A circuit with operations to perform before the Rabi."""
        circ = QuantumCircuit(1)
        circ.x(0)
        return circ



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Fine drag calibration experiment."""

from typing import Dict, Optional, Sequence
import numpy as np

from qiskit.circuit import Gate, QuantumCircuit
from qiskit.providers.backend import Backend
from qiskit.pulse import Play

from qiskit_experiments.exceptions import CalibrationError
from qiskit_experiments.framework import ExperimentData, Options
from qiskit_experiments.calibration_management import (
    BaseCalibrationExperiment,
    Calibrations,
)
from qiskit_experiments.calibration_management.update_library import BaseUpdater
from qiskit_experiments.library.characterization.fine_drag import FineDrag


class FineDragCal(BaseCalibrationExperiment, FineDrag):
    """A calibration version of the fine DRAG experiment."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        schedule_name: str,
        backend: Optional[Backend] = None,
        cal_parameter_name: Optional[str] = "β",
        auto_update: bool = True,
    ):
        r"""See class :class:`.FineDrag` for details.

        Note that this class implicitly assumes that the target angle of the gate
        is :math:`\pi` as seen from the default experiment options.

        Args:
            physical_qubits: Sequence containing the qubit for which to run the
                fine drag calibration.
            calibrations: The calibrations instance with the schedules.
            schedule_name: The name of the schedule to calibrate.
            backend: Optional, the backend to run the experiment on.
            cal_parameter_name: The name of the parameter in the schedule to update.
            auto_update: Whether or not to automatically update the calibrations. By
                default this variable is set to True.
        """
        super().__init__(
            calibrations,
            physical_qubits,
            Gate(name=schedule_name, num_qubits=1, params=[]),
            schedule_name=schedule_name,
            backend=backend,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            target_angle (float): The target rotation angle of the gate being calibrated.
                This value is needed for the update rule.
        """
        options = super()._default_experiment_options()
        options.update_options(target_angle=np.pi)
        return options

    def _metadata(self) -> Dict[str, any]:
        """Add metadata to the experiment data making it more self contained.

        The following keys are added to each experiment's metadata:
            cal_param_value: The value of the drag parameter. This value together with
                the fit result will be used to find the new value of the drag parameter.
            cal_param_name: The name of the parameter in the calibrations.
            cal_schedule: The name of the schedule in the calibrations.
            target_angle: The target angle of the gate.
            cal_group: The calibration group to which the parameter belongs.
        """
        metadata = super()._metadata()
        metadata["target_angle"] = self.experiment_options.target_angle
        metadata["cal_param_value"] = self._cals.get_parameter_value(
            self._param_name,
            self.physical_qubits,
            self._sched_name,
            group=self.experiment_options.group,
        )

        return metadata

    def _attach_calibrations(self, circuit: QuantumCircuit):
        """Attach the calibrations to the circuit."""
        schedule = self._cals.get_schedule(self._sched_name, self.physical_qubits)
        circuit.add_calibration(self._sched_name, self.physical_qubits, schedule)
        # FineDrag always uses sx so attach it if it is not sched_name
        if self._sched_name != "sx":
            schedule = self._cals.get_schedule("sx", self.physical_qubits)
            circuit.add_calibration("sx", self.physical_qubits, schedule)

    def update_calibrations(self, experiment_data: ExperimentData):
        """Update the drag parameter of the pulse in the calibrations."""

        result_index = self.experiment_options.result_index
        group = experiment_data.metadata["cal_group"]
        target_angle = experiment_data.metadata["target_angle"]
        qubits = experiment_data.metadata["physical_qubits"]

        schedule = self._cals.get_schedule(self._sched_name, qubits)

        # Obtain sigma as it is needed for the fine DRAG update rule.
        sigmas = []
        for block in schedule.blocks:
            if isinstance(block, Play) and hasattr(block.pulse, "sigma"):
                sigmas.append(getattr(block.pulse, "sigma"))

        if len(set(sigmas)) != 1:
            raise CalibrationError(
                "Cannot run fine Drag calibration on a schedule with multiple values of sigma."
            )

        if len(sigmas) == 0:
            raise CalibrationError(f"Could not infer sigma from {schedule}.")

        d_theta = BaseUpdater.get_value(experiment_data, "d_theta", result_index)

        # See the documentation in fine_drag.py for the derivation of this rule.
        d_beta = -np.sqrt(np.pi) * d_theta * sigmas[0] / target_angle**2
        old_beta = experiment_data.metadata["cal_param_value"]
        new_beta = old_beta + d_beta

        BaseUpdater.add_parameter_value(
            self._cals, experiment_data, new_beta, self._param_name, schedule, group
        )


class FineXDragCal(FineDragCal):
    """Fine DRAG calibration of X gate."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        backend: Optional[Backend] = None,
        cal_parameter_name: Optional[str] = "β",
        auto_update: bool = True,
    ):
        r"""see class :class:`.FineDrag` for details.

        Args:
            physical_qubits: Sequence containing the qubit for which to run the
                fine drag calibration.
            calibrations: The calibrations instance with the schedules.
            backend: Optional, the backend to run the experiment on.
            cal_parameter_name: The name of the parameter in the schedule to update.
            auto_update: Whether or not to automatically update the calibrations. By
                default this variable is set to True.
        """
        super().__init__(
            physical_qubits,
            calibrations,
            schedule_name="x",
            backend=backend,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )


class FineSXDragCal(FineDragCal):
    """Fine DRAG calibration of X gate."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        backend: Optional[Backend] = None,
        cal_parameter_name: Optional[str] = "β",
        auto_update: bool = True,
    ):
        r"""see class :class:`.FineDrag` for details.

        Args:
            physical_qubits: Sequence containing the qubit for which to run the
                fine drag calibration.
            calibrations: The calibrations instance with the schedules.
            backend: Optional, the backend to run the experiment on.
            cal_parameter_name: The name of the parameter in the schedule to update.
            auto_update: Whether or not to automatically update the calibrations. By
                default this variable is set to True.
        """
        super().__init__(
            physical_qubits,
            calibrations,
            schedule_name="sx",
            backend=backend,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options."""
        options = super()._default_experiment_options()
        options.target_angle = np.pi / 2
        return options



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Data class for parameter values."""

from dataclasses import dataclass
from datetime import datetime
from typing import Union
import warnings

from qiskit_experiments.exceptions import CalibrationError


@dataclass
class ParameterValue:
    """A data class to store parameter values."""

    # Value assumed by the parameter
    value: Union[int, float, complex] = None

    # Data time when the value of the parameter was generated
    date_time: datetime = datetime.fromtimestamp(0)

    # A bool indicating if the parameter is valid
    valid: bool = True

    # The experiment from which the value of this parameter was generated.
    exp_id: str = None

    # The group of calibrations to which this parameter belongs
    group: str = "default"

    def __post_init__(self):
        """
        Ensure that the variables in self have the proper types. This allows
        us to give strings to self.__init__ as input which is useful when loading
        serialized parameter values.
        """
        if isinstance(self.valid, str):
            if self.valid == "True":
                self.valid = True
            else:
                self.valid = False

        if isinstance(self.value, str):
            self.value = self._validated_value(self.value)

        if isinstance(self.date_time, str):
            base_fmt = "%Y-%m-%d %H:%M:%S.%f"
            zone_fmts = ["%z", "", "Z"]
            for time_zone in zone_fmts:
                date_format = base_fmt + time_zone
                try:
                    self.date_time = datetime.strptime(self.date_time, date_format)
                    break
                except ValueError:
                    pass
            else:
                formats = list(base_fmt + zone for zone in zone_fmts)
                raise CalibrationError(
                    f"Cannot parse {self.date_time} in either of {formats} formats."
                )

        self.date_time = self.date_time.astimezone()

        if isinstance(self.value, complex):
            warnings.warn(
                "Support of complex parameters is now pending deprecation, following the"
                "same transition in Qiskit's Pulse module."
                "The main use of complex parameters was the complex amplitude in SymbolicPulse"
                "instances. This use could be removed by converting the pulses to the"
                "ScalableSymbolicPulse class which uses two floats (amp,angle) for the"
                "complex amplitude.",
                PendingDeprecationWarning,
            )

        if not isinstance(self.value, (int, float, complex)):
            raise CalibrationError(f"Values {self.value} must be int, float or complex.")

        if not isinstance(self.date_time, datetime):
            raise CalibrationError(f"Datetime {self.date_time} must be a datetime.")

        if not isinstance(self.valid, bool):
            raise CalibrationError(f"Valid {self.valid} is not a boolean.")

        if self.exp_id and not isinstance(self.exp_id, str):
            raise CalibrationError(f"Experiment id {self.exp_id} is not a string.")

        if not isinstance(self.group, str):
            raise CalibrationError(f"Group {self.group} is not a string.")

    @staticmethod
    def _validated_value(value: str) -> Union[int, float, complex]:
        """
        Convert the string representation of value to the correct type.

        Args:
            value: The string to convert to either an int, float, or complex.

        Returns:
            value converted to either int, float, or complex.

        Raises:
            CalibrationError: If the conversion fails.
        """
        try:
            return int(value)
        except ValueError:
            pass

        try:
            return float(value)
        except ValueError:
            pass

        try:
            return complex(value)
        except ValueError as val_err:
            raise CalibrationError(
                f"Could not convert {value} to int, float, or complex."
            ) from val_err



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Discriminators that wrap SKLearn."""

from typing import Any, List, Dict, TYPE_CHECKING

from qiskit_experiments.data_processing.discriminator import BaseDiscriminator
from qiskit_experiments.framework.package_deps import HAS_SKLEARN

if TYPE_CHECKING:
    from sklearn.discriminant_analysis import (
        LinearDiscriminantAnalysis,
        QuadraticDiscriminantAnalysis,
    )


class SkLDA(BaseDiscriminator):
    """A wrapper for the scikit-learn linear discriminant analysis.

    .. note::
        This class requires that scikit-learn is installed.
    """

    def __init__(self, lda: "LinearDiscriminantAnalysis"):
        """
        Args:
            lda: The sklearn linear discriminant analysis. This may be a trained or an
                untrained discriminator.

        Raises:
            DataProcessorError: If SKlearn could not be imported.
        """
        self._lda = lda
        self.attributes = [
            "coef_",
            "intercept_",
            "covariance_",
            "explained_variance_ratio_",
            "means_",
            "priors_",
            "scalings_",
            "xbar_",
            "classes_",
            "n_features_in_",
            "feature_names_in_",
        ]

    @property
    def discriminator(self) -> Any:
        """Return then SKLearn object."""
        return self._lda

    def is_trained(self) -> bool:
        """Return True if the discriminator has been trained on data."""
        return not getattr(self._lda, "classes_", None) is None

    def predict(self, data: List):
        """Wrap the predict method of the LDA."""
        return self._lda.predict(data)

    def fit(self, data: List, labels: List):
        """Fit the LDA.

        Args:
            data: The independent data.
            labels: The labels corresponding to data.
        """
        self._lda.fit(data, labels)

    def config(self) -> Dict[str, Any]:
        """Return the configuration of the LDA."""
        attr_conf = {attr: getattr(self._lda, attr, None) for attr in self.attributes}
        return {"params": self._lda.get_params(), "attributes": attr_conf}

    @classmethod
    @HAS_SKLEARN.require_in_call
    def from_config(cls, config: Dict[str, Any]) -> "SkLDA":
        """Deserialize from an object."""
        from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

        lda = LinearDiscriminantAnalysis()
        lda.set_params(**config["params"])

        for name, value in config["attributes"].items():
            if value is not None:
                setattr(lda, name, value)

        return SkLDA(lda)


class SkQDA(BaseDiscriminator):
    """A wrapper for the SKlearn quadratic discriminant analysis.

    .. note::
        This class requires that scikit-learn is installed.
    """

    def __init__(self, qda: "QuadraticDiscriminantAnalysis"):
        """
        Args:
            qda: The sklearn quadratic discriminant analysis. This may be a trained or an
                untrained discriminator.

        Raises:
            DataProcessorError: If SKlearn could not be imported.
        """
        self._qda = qda
        self.attributes = [
            "coef_",
            "intercept_",
            "covariance_",
            "explained_variance_ratio_",
            "means_",
            "priors_",
            "scalings_",
            "xbar_",
            "classes_",
            "n_features_in_",
            "feature_names_in_",
            "rotations_",
        ]

    @property
    def discriminator(self) -> Any:
        """Return then SKLearn object."""
        return self._qda

    def is_trained(self) -> bool:
        """Return True if the discriminator has been trained on data."""
        return not getattr(self._qda, "classes_", None) is None

    def predict(self, data: List):
        """Wrap the predict method of the QDA."""
        return self._qda.predict(data)

    def fit(self, data: List, labels: List):
        """Fit the QDA.

        Args:
            data: The independent data.
            labels: The labels corresponding to data.
        """
        self._qda.fit(data, labels)

    def config(self) -> Dict[str, Any]:
        """Return the configuration of the QDA."""
        attr_conf = {attr: getattr(self._qda, attr, None) for attr in self.attributes}
        return {"params": self._qda.get_params(), "attributes": attr_conf}

    @classmethod
    @HAS_SKLEARN.require_in_call
    def from_config(cls, config: Dict[str, Any]) -> "SkQDA":
        """Deserialize from an object."""
        from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis

        qda = QuadraticDiscriminantAnalysis()
        qda.set_params(**config["params"])

        for name, value in config["attributes"].items():
            if value is not None:
                setattr(qda, name, value)

        return SkQDA(qda)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test utility functions."""

import uuid
from typing import Optional, Dict
from datetime import datetime, timezone

from qiskit.providers.job import JobV1 as Job
from qiskit.providers.jobstatus import JobStatus
from qiskit.providers.backend import BackendV1 as Backend
from qiskit.result import Result


class FakeJob(Job):
    """Fake job."""

    def __init__(self, backend: Backend, result: Optional[Result] = None):
        """Initialize FakeJob."""
        if result:
            job_id = result.job_id
        else:
            job_id = uuid.uuid4().hex
        super().__init__(backend, job_id)
        self._result = result

    def result(self):
        """Return job result."""
        return self._result

    def submit(self):
        """Submit the job to the backend for execution."""
        pass

    @staticmethod
    def time_per_step() -> Dict[str, datetime]:
        """Return the completion time."""
        return {"COMPLETED": datetime.now(timezone.utc)}

    def status(self) -> JobStatus:
        """Return the status of the job, among the values of ``JobStatus``."""
        if self._result:
            return JobStatus.DONE
        return JobStatus.RUNNING



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Fine amplitude calibration experiment."""

from typing import Dict, Optional, Sequence
import numpy as np

from qiskit.circuit import Gate, QuantumCircuit
from qiskit.providers.backend import Backend

from qiskit_experiments.calibration_management import (
    BaseCalibrationExperiment,
    Calibrations,
)
from qiskit_experiments.library.characterization import FineAmplitude
from qiskit_experiments.framework import ExperimentData, Options
from qiskit_experiments.calibration_management.update_library import BaseUpdater


class FineAmplitudeCal(BaseCalibrationExperiment, FineAmplitude):
    r"""A calibration version of the :class:`.FineAmplitude` experiment.

    # section: overview

        :class:`FineAmplitudeCal` is a subclass of :class:`.FineAmplitude`. In the calibration
        experiment the circuits that are run have a custom gate with the pulse schedule attached
        to it through the calibrations.

    """

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        schedule_name: str,
        backend: Optional[Backend] = None,
        cal_parameter_name: Optional[str] = "amp",
        auto_update: bool = True,
        gate: Optional[Gate] = None,
        measurement_qubits: Sequence[int] = None,
    ):
        """See class :class:`FineAmplitude` for details.

        Args:
            physical_qubits: Sequence containing the qubit(s) for which to run
                the fine amplitude calibration. This can be a pair of qubits
                which correspond to control and target qubit.
            calibrations: The calibrations instance with the schedules.
            schedule_name: The name of the schedule to calibrate.
            backend: Optional, the backend to run the experiment on.
            cal_parameter_name: The name of the parameter in the schedule to update.
            auto_update: Whether or not to automatically update the calibrations. By
                default this variable is set to True.
            gate: The gate to repeat in the quantum circuit. If this argument
                is None (the default), then the gate is built from the schedule name.
            measurement_qubits: The qubits in the given physical qubits that need to
                be measured.
        """
        gate = gate or Gate(name=schedule_name, num_qubits=len(physical_qubits), params=[])

        super().__init__(
            calibrations,
            physical_qubits,
            gate,
            schedule_name=schedule_name,
            backend=backend,
            measurement_qubits=measurement_qubits,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )

    @classmethod
    def _default_experiment_options(cls):
        """Default values for the fine amplitude calibration experiment.

        Experiment Options:
            target_angle (float): The target angle of the pulse.
        """
        options = super()._default_experiment_options()
        options.target_angle = np.pi
        return options

    def _metadata(self) -> Dict[str, any]:
        """Add metadata to the experiment data making it more self contained.

        The following keys are added to each experiment's metadata:
            cal_param_value: The value of the pulse amplitude. This value together with
                the fit result will be used to find the new value of the pulse amplitude.
            cal_param_name: The name of the parameter in the calibrations.
            cal_schedule: The name of the schedule in the calibrations.
            target_angle: The target angle of the gate.
            cal_group: The calibration group to which the parameter belongs.
        """
        metadata = super()._metadata()
        metadata["target_angle"] = self.experiment_options.target_angle
        metadata["cal_param_value"] = self._cals.get_parameter_value(
            self._param_name,
            self.physical_qubits,
            self._sched_name,
            group=self.experiment_options.group,
        )

        return metadata

    def _attach_calibrations(self, circuit: QuantumCircuit):
        """Adds the calibrations to the transpiled circuits."""
        for gate in ["x", "sx"]:
            schedule = self._cals.get_schedule(gate, self.physical_qubits)
            circuit.add_calibration(gate, self.physical_qubits, schedule)

    def update_calibrations(self, experiment_data: ExperimentData):
        r"""Update the amplitude of the pulse in the calibrations.

        The update rule of this experiment is

        .. math::

            A \to A \frac{\theta_\text{target}}{\theta_\text{target} + {\rm d}\theta}

        Where :math:`A` is the amplitude of the pulse before the update.

        Args:
            experiment_data: The experiment data from which to extract the measured over/under
                rotation used to adjust the amplitude.
        """

        result_index = self.experiment_options.result_index
        group = experiment_data.metadata["cal_group"]
        target_angle = experiment_data.metadata["target_angle"]
        prev_amp = experiment_data.metadata["cal_param_value"]

        # Protect against cases where the complex amplitude was converted to a list.
        if isinstance(prev_amp, list) and len(prev_amp) == 2:
            prev_amp = prev_amp[0] + 1.0j * prev_amp[1]

        d_theta = BaseUpdater.get_value(experiment_data, "d_theta", result_index)

        BaseUpdater.add_parameter_value(
            self._cals,
            experiment_data,
            prev_amp * target_angle / (target_angle + d_theta),
            self._param_name,
            self._sched_name,
            group,
        )


class FineXAmplitudeCal(FineAmplitudeCal):
    """A calibration experiment to calibrate the amplitude of the X schedule."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        schedule_name: str,
        backend: Optional[Backend] = None,
        cal_parameter_name: Optional[str] = "amp",
        auto_update: bool = True,
    ):
        super().__init__(
            physical_qubits,
            calibrations,
            schedule_name,
            backend=backend,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )
        self.analysis.set_options(
            fixed_parameters={
                "angle_per_gate": np.pi,
                "phase_offset": np.pi / 2,
            }
        )

    @classmethod
    def _default_transpile_options(cls):
        """Default transpile options.

        Transpile Options:
            basis_gates (list(str)): A list of basis gates needed for this experiment.
                The schedules for these basis gates will be provided by the instruction
                schedule map from the calibrations.
        """
        options = super()._default_transpile_options()
        options.basis_gates = ["x", "sx"]

        return options

    def _pre_circuit(self, num_clbits: int) -> QuantumCircuit:
        """The preparation circuit is an sx gate to move to the equator of the Bloch sphere."""
        circuit = QuantumCircuit(self.num_qubits, num_clbits)
        circuit.sx(0)
        return circuit


class FineSXAmplitudeCal(FineAmplitudeCal):
    """A calibration experiment to calibrate the amplitude of the SX schedule."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        schedule_name: str,
        backend: Optional[Backend] = None,
        cal_parameter_name: Optional[str] = "amp",
        auto_update: bool = True,
    ):
        super().__init__(
            physical_qubits,
            calibrations,
            schedule_name,
            backend=backend,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )
        self.analysis.set_options(
            fixed_parameters={
                "angle_per_gate": np.pi / 2,
                "phase_offset": np.pi,
            }
        )

    @classmethod
    def _default_experiment_options(cls) -> Options:
        r"""Default values for the fine amplitude experiment.

        Experiment Options:
            add_sx (bool): This option is False by default when calibrating gates with a target
                angle per gate of :math:`\pi/2` as this increases the sensitivity of the
                experiment.
            add_xp_circuit (bool): This option is False by default when calibrating gates with
                a target angle per gate of :math:`\pi/2`.
            repetitions (List[int]): By default the repetitions take on odd numbers for
                :math:`\pi/2` target angles as this ideally prepares states on the equator of
                the Bloch sphere. Note that the repetitions include two repetitions which
                plays the same role as including a circuit with an X gate.
            target_angle (float): The target angle per gate.
        """
        options = super()._default_experiment_options()
        options.add_cal_circuits = False
        options.repetitions = [0, 1, 2, 3, 5, 7, 9, 11, 13, 15, 17, 21, 23, 25]
        options.target_angle = np.pi / 2
        return options



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Types used by the calibration module."""

from typing import NamedTuple, Tuple, Union
from collections import namedtuple

from qiskit.circuit import ParameterExpression


ParameterKey = namedtuple("ParameterKey", ["parameter", "qubits", "schedule"])
ParameterValueType = Union[ParameterExpression, float, int, complex]


class DefaultCalValue(NamedTuple):
    """Defines the structure of a default value."""

    value: Union[float, int, complex]
    parameter: str
    qubits: Tuple
    schedule_name: str


class ScheduleKey(NamedTuple):
    """Defines the structure of a key to find a schedule."""

    schedule: str  # Name of the schedule
    qubits: Tuple  # Qubits the schedule acts on

    def __repr__(self):
        return f"{self.schedule}::{self.qubits}"

    @classmethod
    def from_repr(cls, rep_str: str) -> "ScheduleKey":
        """Construct a key form its representation as a string."""
        name, qubits = rep_str.split("::")
        qubits = tuple(int(qubit) for qubit in qubits.strip("( )").split(",") if qubit != "")
        return ScheduleKey(name, qubits)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"The Qiskit Experiments setup file."

import os
from setuptools import setup, find_packages

with open("requirements.txt", encoding="utf-8") as f:
    REQUIREMENTS = f.read().splitlines()

with open("requirements-extras.txt", encoding="utf-8") as f:
    EXTRAS = f.read().splitlines()

version_path = os.path.abspath(
    os.path.join(os.path.join(os.path.dirname(__file__), "qiskit_experiments"), "VERSION.txt")
)
with open(version_path, "r", encoding="utf-8") as fd:
    version = fd.read().rstrip()

# Read long description from README.
README_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")
with open(README_PATH, encoding="utf-8") as readme_file:
    README = readme_file.read()

setup(
    name="qiskit-experiments",
    version=version,
    description="Software for developing quantum computing programs",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Qiskit-Community/qiskit-experiments",
    author="Qiskit Development Team",
    author_email="qiskit@us.ibm.com",
    license="Apache 2.0",
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
    ],
    keywords="qiskit sdk quantum",
    packages=find_packages(exclude=["test*"]),
    install_requires=REQUIREMENTS,
    extras_require={"extras": EXTRAS},
    include_package_data=True,
    python_requires=">=3.9",
    project_urls={
        "Bug Tracker": "https://github.com/Qiskit-Community/qiskit-experiments/issues",
        "Documentation": "https://qiskit-community.github.io/qiskit-experiments",
        "Source Code": "https://github.com/Qiskit-Community/qiskit-experiments",
    },
    zip_safe=False,
    entry_points={
        "qiskit.synthesis": [
            "clifford.rb_default = qiskit_experiments.library.randomized_benchmarking.clifford_synthesis:RBDefaultCliffordSynthesis",
        ],
    },
)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Settings mixing class
"""

import inspect
from collections import OrderedDict
from functools import wraps


class StoreInitArgs:
    """Class mixing for storing class and subclass instance init args.

    This mixin adds a ``__new__`` method that stores the values of args
    and kwargs passed to the class instances ``__init__`` method.
    These are stored as ordered dicts under attributes ``__init_args__`
    and ``__init_kwargs__`` respectively.

    .. note::

        This mixin is intended as a mixing for base classes so that when
        creating subclasses, those subclasses can inherit the logic for
        saving and returning settings.

        Note that there is small performance overhead to initializing classes
        with this mixin so it should not be used for adding settings to all
        classes without consideration.
    """

    def __new__(cls, *args, **kwargs):
        # This method automatically stores all arg and kwargs from subclass
        # init methods
        spec = inspect.getfullargspec(cls.__init__)
        ord_args = OrderedDict()
        ord_kwargs = OrderedDict()

        # Parse spec args
        defaults = spec.defaults or []
        if defaults:
            size = len(spec.args) - len(spec.defaults)
            init_args = spec.args[1:size]
            init_kwargs = spec.args[size:]
        else:
            init_args = spec.args[1:]
            init_kwargs = []

        # Initialize defaults to preserve correct arg order
        num_args = len(init_args)
        ord_args.update(zip(init_args, num_args * [None]))
        ord_kwargs.update(zip(init_kwargs, defaults))

        if init_args and args:
            # Add named args
            ord_args.update(zip(init_args, args))
        if init_kwargs and args:
            # Update non-default values
            ord_kwargs.update(zip(init_kwargs, args[num_args:]))

        # Parse variadic args
        if spec.varargs:
            num_varargs = len(args) - num_args
            ord_args.update(
                ((f"{spec.varargs}[{i}]", args[num_args + i]) for i in range(num_varargs))
            )

        # Add defaults for kwonly args
        for kwarg in spec.kwonlyargs:
            if kwarg not in ord_kwargs:
                ord_kwargs[kwarg] = spec.kwonlydefaults.get(kwarg, None)

        # Parse kwargs
        for arg, argval in kwargs.items():
            if arg in init_args:
                ord_args[arg] = argval
            else:
                ord_kwargs[arg] = argval

        # pylint: disable = attribute-defined-outside-init
        instance = super().__new__(cls)
        instance.__init_args__ = ord_args
        instance.__init_kwargs__ = ord_kwargs
        return instance

    def __init_subclass__(cls, **kwargs):
        # This method fixes class documentations for subclass
        # that inherit the base class new method
        super().__init_subclass__(**kwargs)

        # Copy the doc string and annotation from the subclasses
        # init method to its new method to override base class
        # __new__ documentation
        @wraps(cls.__init__, assigned=("__annotations__",))
        def __new__(sub_cls, *args, **kwargs):
            return super(cls, sub_cls).__new__(sub_cls, *args, **kwargs)

        # Monkey patch the subclass new method with the method with
        # fixed documentation annotations
        cls.__new__ = __new__



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A helper module to save calibration data to local storage.

.. warning::

    This module is expected to be internal and is not intended as a stable user-facing API.

.. note::

    Because a locally saved :class:`.Calibrations` instance may not conform to the
    data model of the latest Qiskit Experiments implementation, the calibration loader
    must be aware of the data model version.
    CalibrationModel classes representing the data model must have
    the version suffix, e.g., `CalibrationModelV1` and the `schema_version` field.
    This helps the loader to raise user-friendly error rather than being crashed by
    incompatible data, and possibly to dispatch the loader function based on the version number.

    When a developer refactors the :class:`.Calibrations` class to a new data model,
    the developer must also define a corresponding CalibrationModel class with new version number.
    Existing CalibrationModel classes should be preserved for backward compatibility.


.. note::

    We don't guarantee the portability of stored data across different Qiskit Experiments
    versions. We allow the calibration loader to raise an error for non-supported
    data models.

"""

from dataclasses import dataclass, field, asdict
from datetime import datetime
from typing import List, Dict, Any

from qiskit import QuantumCircuit
from qiskit.circuit import Instruction
from qiskit.pulse import ScheduleBlock

from .calibrations import Calibrations
from .control_channel_map import ControlChannelMap
from .parameter_value import ParameterValue


@dataclass(frozen=True)
class ParameterModelV1:
    """A data schema of a single calibrated parameter.

    .. note::
        This is intentionally agnostic to the data structure of
        Qiskit Experiments Calibrations for portability.

    """

    param_name: str
    """Name of the parameter."""

    qubits: List[int]
    """List of associated qubits."""

    schedule: str = ""
    """Associated schedule name."""

    value: float = 0.0
    """Parameter value."""

    datetime: datetime = None
    """A date time at which this value is obtained."""

    valid: bool = True
    """If this parameter is valid."""

    exp_id: str = ""
    """Associated experiment ID which is used to obtain this value."""

    group: str = ""
    """Name of calibration group in which this calibration parameter belongs to."""

    @classmethod
    def apply_schema(cls, data: Dict[str, Any]):
        """Consume dictionary and returns canonical data model."""
        return ParameterModelV1(**data)


@dataclass
class CalibrationModelV1:
    """A data schema to represent instances of Calibrations.

    .. note::
        This is intentionally agnostic to the data structure of
        Qiskit Experiments Calibrations for portability.

    """

    backend_name: str
    """Name of the backend."""

    backend_version: str
    """Version of the backend."""

    device_coupling_graph: List[List[int]]
    """Qubit coupling graph of the device."""

    control_channel_map: ControlChannelMap
    """Mapping of ControlChannel to qubit index."""

    schedules: List[ScheduleBlock] = field(default_factory=list)
    """Template schedules. It must contain the metadata for qubits and num_qubits."""

    parameters: List[ParameterModelV1] = field(default_factory=list)
    """List of calibrated pulse parameters."""

    schedule_free_parameters: QuantumCircuit = field(default_factory=lambda: QuantumCircuit(1))
    """Placeholder circuit for parameters not associated with a schedule

    The circuit contains placeholder instructions which have the Parameter
    objects attached and operate on the qubits that the parameter is associated
    with in the calibrations.
    """

    schema_version: str = "1.0"
    """Version of this data model. This must be static."""

    @classmethod
    def apply_schema(cls, data: Dict[str, Any]):
        """Consume dictionary and returns canonical data model."""
        in_data = {}
        for key, value in data.items():
            if key == "parameters":
                value = list(map(ParameterModelV1.apply_schema, value))
            in_data[key] = value
        return CalibrationModelV1(**in_data)


def calibrations_to_dict(
    cals: Calibrations,
    most_recent_only: bool = True,
) -> Dict[str, Any]:
    """A helper function to convert calibration data into dictionary.

    Args:
        cals: A calibration instance to save.
        most_recent_only: Set True to save calibration parameters with most recent time stamps.

    Returns:
        Canonical calibration data in dictionary format.
    """
    schedules = getattr(cals, "_schedules")
    num_qubits = getattr(cals, "_schedules_qubits")
    parameters = getattr(cals, "_params")
    if most_recent_only:
        # Get values with most recent time stamps.
        parameters = {k: [max(parameters[k], key=lambda x: x.date_time)] for k in parameters}

    data_entries = []
    for param_key, param_values in parameters.items():
        for param_value in param_values:
            entry = ParameterModelV1(
                param_name=param_key.parameter,
                qubits=param_key.qubits,
                schedule=param_key.schedule,
                value=param_value.value,
                datetime=param_value.date_time,
                valid=param_value.valid,
                exp_id=param_value.exp_id,
                group=param_value.group,
            )
            data_entries.append(entry)

    sched_entries = []
    for sched_key, sched_obj in schedules.items():
        if "qubits" not in sched_obj.metadata or "num_qubits" not in sched_obj.metadata:
            qubit_metadata = {
                "qubits": sched_key.qubits,
                "num_qubits": num_qubits[sched_key],
            }
            sched_obj.metadata.update(qubit_metadata)
        sched_entries.append(sched_obj)

    max_qubit = max(
        (max(k.qubits or (0,)) for k in cals._parameter_map if k.schedule is None),
        default=0,
    )
    schedule_free_parameters = QuantumCircuit(max_qubit + 1)
    for sched_key, param in cals._parameter_map.items():
        if sched_key.schedule is None:
            schedule_free_parameters.append(
                Instruction("parameter_container", len(sched_key.qubits), 0, [param]),
                sched_key.qubits,
            )

    model = CalibrationModelV1(
        backend_name=cals.backend_name,
        backend_version=cals.backend_version,
        device_coupling_graph=getattr(cals, "_coupling_map"),
        control_channel_map=ControlChannelMap(getattr(cals, "_control_channel_map")),
        schedules=sched_entries,
        parameters=data_entries,
        schedule_free_parameters=schedule_free_parameters,
    )

    return asdict(model)


def calibrations_from_dict(
    cal_data: Dict[str, Any],
) -> Calibrations:
    """A helper function to build calibration instance from canonical dictionary.

    Args:
        cal_data: Calibration data dictionary which is formatted according to the
            predefined data schema provided by Qiskit Experiments.
            This formatting is implicitly performed when the calibration data is
            dumped into dictionary with the :func:`calibrations_to_dict` function.

    Returns:
        Calibration instance.

    Raises:
        ValueError: When input data model version is not supported.
        KeyError: When input data model doesn't conform to data schema.
    """
    # Apply schema for data field validation
    try:
        version = cal_data["schema_version"]
        if version == "1.0":
            model = CalibrationModelV1.apply_schema(cal_data)
        else:
            raise ValueError(
                f"Loading calibration data with schema version {version} is no longer supported. "
                "Use the same version of Qiskit Experiments at the time of saving."
            )
    except (KeyError, TypeError) as ex:
        raise KeyError(
            "Loaded data doesn't match with the defined data schema. "
            "Check if this object is dumped from the Calibrations instance."
        ) from ex

    # This can dispatch loading mechanism depending on schema version
    cals = Calibrations(
        coupling_map=model.device_coupling_graph,
        control_channel_map=model.control_channel_map.chan_map,
        backend_name=model.backend_name,
        backend_version=model.backend_version,
    )

    # Add schedules
    for sched in model.schedules:
        qubits = sched.metadata.pop("qubits", tuple())
        num_qubits = sched.metadata.pop("num_qubits", None)
        cals.add_schedule(
            schedule=sched,
            qubits=qubits if qubits and len(qubits) != 0 else None,
            num_qubits=num_qubits,
        )

    # Add parameters
    for param in model.parameters:
        param_value = ParameterValue(
            value=param.value,
            date_time=param.datetime,
            valid=param.valid,
            exp_id=param.exp_id,
            group=param.group,
        )
        cals.add_parameter_value(
            value=param_value,
            param=param.param_name,
            qubits=tuple(param.qubits),
            schedule=param.schedule,
            update_inst_map=False,
        )

    for instruction in model.schedule_free_parameters.data:
        # For some reason, pylint thinks the items in data are tuples instead
        # of CircuitInstruction. Remove the following line if it ever stops
        # thinking that:
        # pylint: disable=no-member
        for param in instruction.operation.params:
            cals._register_parameter(param, instruction.qubits)

    cals.update_inst_map()

    return cals



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Multi state discrimination experiment."""

import warnings
from typing import Dict, List, Optional, Sequence

from qiskit import QuantumCircuit
from qiskit.circuit import Gate
from qiskit.providers import Backend
from qiskit.providers.options import Options
from qiskit.pulse import ScheduleBlock
from qiskit.qobj.utils import MeasLevel, MeasReturnType
from qiskit_experiments.framework import BaseExperiment
from qiskit_experiments.library.characterization import MultiStateDiscriminationAnalysis


class MultiStateDiscrimination(BaseExperiment):
    r"""An experiment that discriminates between the first :math:`n` energy states.

    # section: overview

        The experiment creates :math:`n` circuits that prepare, respectively, the energy states
        :math:`|0\rangle,\cdots,|n-1\rangle`. For, e.g., :math:`n=4` the circuits are of the form

        .. parsed-literal::

            Circuit preparing :math:`|0\rangle`

                       ░ ┌─┐
                   q: ─░─┤M├
                       ░ └╥┘
                meas: ════╩═

            ...

            Circuit preparing :math:`|3\rangle`

                      ┌───┐┌─────┐┌─────┐ ░ ┌─┐
                   q: ┤ X ├┤ x12 ├┤ x23 ├─░─┤M├
                      └───┘└─────┘└─────┘ ░ └╥┘
                meas: ═══════════════════════╩═

    # section: analysis_ref
        :class:`MultiStateDiscriminationAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            # Temporary workaround for missing support in Qiskit and qiskit-ibm-runtime
            from qiskit_experiments.test.patching import patch_sampler_test_support
            patch_sampler_test_support()

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9,-.25e9, 1e9, 0.8e9, 1e4, noise=False, seed=199)

        .. jupyter-execute::

            from qiskit_experiments.library.characterization import MultiStateDiscrimination

            exp=MultiStateDiscrimination((0,), backend=backend)

            exp_data=exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)

    # section: reference
        `Qiskit Textbook\
        <https://github.com/Qiskit/textbook/blob/main/notebooks/quantum-hardware-pulses/accessing_higher_energy_states.ipynb>`_

    """

    @classmethod
    def _default_run_options(cls) -> Options:
        """Default option values for the experiment :meth:`run` method."""
        options = super()._default_run_options()

        options.meas_level = MeasLevel.KERNELED
        options.meas_return = MeasReturnType.SINGLE

        return options

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default values for the number of states if none is given.

        Experiment Options:
            n_states (int): The number of states to discriminate.
            schedules (dict): A dictionary of the schedules for the gates in the experiment. Each key is
                a gate name of the form ``xii+1`` which should implement an x-rotation between level
                ``i`` and ``i+1``.

        """
        options = super()._default_experiment_options()
        options.n_states = 2
        options.schedules = None

        return options

    def __init__(
        self,
        physical_qubits: Sequence[int],
        backend: Optional[Backend] = None,
        n_states: Optional[int] = None,
        schedules: Optional[Dict[str, ScheduleBlock]] = None,
    ):
        """Setup an experiment to prepare different energy states on a given qubit.

        Args:
            physical_qubits: A single-element sequence containing the qubit on which to run the
                experiment.
            backend: Optional, the backend to run the experiment on.
            n_states: The number of energy levels to prepare.
            schedules: The schedules of the x gates between neighboring energy levels.
        """

        super().__init__(
            physical_qubits, analysis=MultiStateDiscriminationAnalysis(), backend=backend
        )

        self.experiment_options.schedules = schedules

        if n_states is not None:
            self.set_experiment_options(n_states=n_states)

    def circuits(self) -> List[QuantumCircuit]:
        """
        Create the circuits for the multi state discrimination experiment.

        Returns:
            A list of circuits preparing the different energy states.
        """
        warnings.warn(
            (
                "Setting pulse schedules for x gates is deprecated  as of "
                "version 0.8 due to the deprecation of Qiskit Pulse. It will be "
                "removed in a future release."
            ),
            DeprecationWarning,
        )
        circuits = []
        for level in range(self.experiment_options.n_states):
            circuit = QuantumCircuit(1)

            # Prepare |1>
            if level >= 1:
                circuit.x(0)

            # Prepare higher energy states
            if level >= 2:
                for idx in range(1, level):
                    gate_name = f"x{idx}{idx + 1}"
                    gate = Gate(name=gate_name, num_qubits=1, params=[])
                    circuit.append(gate, (0,))
                    if self.experiment_options.schedules is not None:
                        circuit.add_calibration(
                            gate_name,
                            self._physical_qubits,
                            self.experiment_options.schedules[gate_name],
                        )

            # label the circuit
            circuit.metadata = {"label": level}

            circuit.measure_all()
            circuits.append(circuit)

        return circuits



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Drawer abstract class."""

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import numpy as np

from qiskit_experiments.framework import Options

from ..style import PlotStyle
from ..utils import ExtentTuple

SeriesName = Union[str, int, float]


class BaseDrawer(ABC):
    """Abstract class for the serializable Qiskit Experiments figure drawer.

    # section: overview

    A drawer may be implemented by different drawer backends such as matplotlib or
    Plotly. Sub-classes that wrap these backends by subclassing :class:`BaseDrawer` must
    implement the following abstract methods.

    .. describe:: initialize_canvas

        This method should implement a protocol to initialize a drawer canvas with user
        input ``axis`` object. Note that ``drawer`` supports visualization of experiment
        results in multiple canvases tiled into N (row) x M (column) inset grids, which
        is specified in the option ``subplots``. By default, this is N=1, M=1 and thus
        no inset grid will be initialized.

        This method should first check the drawer options (:attr:`options`) for the axis
        object and initialize the axis only when it is not provided by the options. Once
        axis is initialized, this is set to the instance member ``self._axis``.

    .. describe:: format_canvas

        This method formats the appearance of the canvas. Typically, it updates axis and
        tick labels. Note that the axis SI unit may be specified in the drawer
        figure_options. In this case, axis numbers should be auto-scaled with the unit
        prefix.

    .. rubric:: Drawing Methods

    .. describe:: scatter

        This method draws scatter points on the canvas, like a scatter-plot, with
        optional error-bars in both the X and Y axes.

    .. describe:: line

        This method plots a line from provided X and Y values.

    .. describe:: filled_y_area

        This method plots a shaped region bounded by upper and lower Y-values. This
        method is typically called with interpolated x and a pair of y values that
        represent the upper and lower bound within certain confidence interval. If this
        is called multiple times, it may be necessary to set the transparency so that
        overlapping regions can be distinguished.

    .. describe:: filled_x_area

        This method plots a shaped region bounded by upper and lower X-values, as a
        function of Y-values. This method is a rotated analogue of
        :meth:`filled_y_area`.

    .. describe:: textbox

        This method draws a text-box on the canvas, which is a rectangular region
        containing some text.

    .. rubric:: Legends

    Legends are generated based off of drawn graphics and their labels or names. These
    are managed by individual drawer subclasses, and generated when the
    :meth:`format_canvas` method is called. Legend entries are created when any drawing
    function is called with ``legend=True``. There are three parameters in drawing
    functions that are relevant to legend generation: ``name``, ``label``, and
    ``legend``. If a user would like the graphics drawn onto a canvas to be used as the
    graphical component of a legend entry; they should set ``legend=True``. The legend
    entry label can be defined in three locations: the ``label`` parameter of drawing
    functions, the ``"label"`` entry in ``series_params``, and the ``name`` parameter of
    drawing functions. These three possible label variables have a search hierarchy
    given by the order in the aforementioned list. If one of the label variables is
    ``None``, the next is used. If all are ``None``, a legend entry is not generated for
    the given series.

    The recommended way to customize the legend entries is as follows:

    1. Set the labels in the ``series_params`` option, keyed on the series names.
    2. Initialize the canvas.
    3. Call relevant drawing methods to create the figure. When calling the drawing
       method that creates the graphic you would like to use in the legend, set
       ``legend=True``. For example, ``drawer.scatter(...,legend=True)`` would use
       the scatter points as the legend graphics for the given series.
    4. Format the canvas and call :meth:`figure` to get the figure.

    .. rubric:: Options and Figure Options

    Drawers have both :attr:`options` and :attr:`figure_options` available to set
    parameters that define how to draw and what is drawn, respectively.
    :class:`BasePlotter` is similar in that it also has ``options`` and
    ``figure_options``. The former contains class-specific variables that define how an
    instance behaves. The latter contains figure-specific variables that typically
    contain values that are drawn on the canvas, such as text. For details on the
    difference between the two sets of options, see the documentation for
    :class:`BasePlotter`.

    .. note::
        If a drawer instance is used with a plotter, then there is the potential for
        any figure option to be overwritten with their value from the plotter. This
        means that the drawer instance would be modified indirectly when the
        :meth:`BasePlotter.figure` method is called. This must be kept in mind when
        creating subclasses of :class:`BaseDrawer`.

    """

    def __init__(self):
        """Create a BaseDrawer instance."""
        # Normal options. Which includes the drawer axis, subplots, and default style.
        self._options = self._default_options()
        # A set of changed options for serialization.
        self._set_options = set()

        # Figure options which are typically updated by a plotter instance. Figure options include the
        # axis labels, figure title, and a custom style instance.
        self._figure_options = self._default_figure_options()
        # A set of changed figure options for serialization.
        self._set_figure_options = set()

        # The initialized axis/axes, set by `initialize_canvas`.
        self._axis = None

    @property
    def options(self) -> Options:
        """Return the drawer options."""
        return self._options

    @property
    def figure_options(self) -> Options:
        """Return the figure options.

        These are typically updated by a plotter instance, and thus may change. It is
        recommended to set figure options in a parent :class:`BasePlotter` instance that
        contains the :class:`BaseDrawer` instance.
        """
        return self._figure_options

    @classmethod
    def _default_options(cls) -> Options:
        """Return default drawer options.

        Options:
            axis (Any): Arbitrary object that can be used as a canvas.
            subplots (Tuple[int, int]): Number of rows and columns when the experimental
                result is drawn in the multiple windows.
            default_style (PlotStyle): The default style for drawer. This must contain
                all required style parameters for :class:`drawer`, as is defined in
                :meth:`PlotStyle.default_style()`. Subclasses can add extra required
                style parameters by overriding :meth:`_default_style`.
        """
        return Options(
            axis=None,
            subplots=(1, 1),
            default_style=cls._default_style(),
        )

    @classmethod
    def _default_style(cls) -> PlotStyle:
        return PlotStyle.default_style()

    @classmethod
    def _default_figure_options(cls) -> Options:
        """Return default figure options.

        Figure Options:
            xlabel (Union[str, List[str]]): X-axis label string of the output figure. If
                there are multiple columns in the canvas, this could be a list of labels.
            ylabel (Union[str, List[str]]): Y-axis label string of the output figure. If
                there are multiple rows in the canvas, this could be a list of labels.
            xlim (Union[Tuple[float, float], List[Tuple[float, float]]): Min and max value
                of the horizontal axis. If not provided, it is automatically scaled based
                on the input data points. If there are multiple columns in the canvas,
                this could be a list of xlims.
            ylim (Union[Tuple[float, float], List[Tuple[float, float]]): Min and max value
                of the vertical axis. If not provided, it is automatically scaled based
                on the input data points. If there are multiple rows in the canvas,
                this could be a list of ylims.
            xval_unit (Union[str, List[str]]): Unit of x values.
                No scaling prefix is needed here as this is controlled by ``xval_unit_scale``.
                If there are multiple columns in the canvas, this could be a list of xval_units.
            yval_unit (Union[str, List[str]]): Unit of y values.
                No scaling prefix is needed here as this is controlled by ``yval_unit_scale``.
                If there are multiple rows in the canvas, this could be a list of yval_units.
            xval_unit_scale (Union[bool, List[bool]]): Whether to add an SI unit prefix to
                ``xval_unit`` if needed. For example, when the x values represent time and
                ``xval_unit="s"``, ``xval_unit_scale=True`` adds an SI unit prefix to
                ``"s"`` based on X values of plotted data. In the output figure, the
                prefix is automatically selected based on the maximum value in this
                axis. If your x values are in [1e-3, 1e-4], they are displayed as [1 ms,
                10 ms]. By default, this option is set to ``True``. If ``False`` is
                provided, the axis numbers will be displayed in the scientific notation.
                If there are multiple columns in the canvas, this could be a list of xval_unit_scale.
            yval_unit_scale (Union[bool, List[bool]]): Whether to add an SI unit prefix to
                ``yval_unit`` if needed. See ``xval_unit_scale`` for details.
                If there are multiple rows in the canvas, this could be a list of yval_unit_scale.
            xscale (str): The scaling of the x-axis, such as ``log`` or ``linear``.
            yscale (str): The scaling of the y-axis, such as ``log`` or ``linear``.
            figure_title (str): Title of the figure. Defaults to None, i.e. nothing is
                shown.
            sharex (bool): Set True to share x-axis ticks among sub-plots.
            sharey (bool): Set True to share y-axis ticks among sub-plots.
            series_params (Dict[str, Dict[str, Any]]): A dictionary of parameters for
                each series. This is keyed on the name for each series. Sub-dictionary
                is expected to have the following three configurations, "canvas",
                "color", "symbol" and "label"; "canvas" is the integer index of axis
                (when multi-canvas plot is set), "color" is the color of the drawn
                graphics, "symbol" is the series marker style for scatter plots, and
                "label" is a user provided series label that appears in the legend.
            custom_style (PlotStyle): The style definition to use when drawing. This
                overwrites style parameters in ``default_style`` in :attr:`options`.
                Defaults to an empty PlotStyle instance (i.e., ``PlotStyle()``).
        """
        options = Options(
            xlabel=None,
            ylabel=None,
            xlim=None,
            ylim=None,
            xval_unit=None,
            yval_unit=None,
            xval_unit_scale=True,
            yval_unit_scale=True,
            xscale=None,
            yscale=None,
            sharex=True,
            sharey=True,
            figure_title=None,
            series_params={},
            custom_style=PlotStyle(),
        )
        options.set_validator("xscale", ["linear", "log", "symlog", "logit", "quadratic", None])
        options.set_validator("yscale", ["linear", "log", "symlog", "logit", "quadratic", None])

        return options

    def set_options(self, **fields):
        """Set the drawer options.

        Args:
            fields: The fields to update the options

        Raises:
            AttributeError: If an unknown options is encountered.
        """
        for field in fields:
            if not hasattr(self._options, field):
                raise AttributeError(
                    f"Options field {field} is not valid for {type(self).__name__}"
                )

        self._options.update_options(**fields)
        self._set_options = self._set_options.union(fields)

    def set_figure_options(self, **fields):
        """Set the figure options.

        Args:
            fields: The fields to update the figure options

        Raises:
            AttributeError: If an unknown figure option is encountered.
        """
        for field in fields:
            if not hasattr(self._figure_options, field):
                raise AttributeError(
                    f"Figure options field {field} is not valid for {type(self).__name__}"
                )
        self._figure_options.update_options(**fields)
        self._set_figure_options = self._set_figure_options.union(fields)

    @property
    def style(self) -> PlotStyle:
        """The combined plot style for this drawer.

        The returned style instance is a combination of :attr:`options.default_style`
        and :attr:`figure_options.custom_style`. Style parameters set in
        ``custom_style`` override those set in ``default_style``. If ``custom_style`` is
        not an instance of :class:`PlotStyle`, the returned style is equivalent to
        ``default_style``.

        Returns:
            The plot style for this drawer.
        """
        if isinstance(self.figure_options.custom_style, PlotStyle):
            return PlotStyle.merge(self.options.default_style, self.figure_options.custom_style)
        return self.options.default_style

    @abstractmethod
    def initialize_canvas(self):
        """Initialize the drawer canvas."""

    @abstractmethod
    def format_canvas(self):
        """Final cleanup for the canvas appearance."""

    def label_for(self, name: Optional[SeriesName], label: Optional[SeriesName]) -> Optional[str]:
        """Get the legend label for the given series, with optional overrides.

        This method determines the legend label for a series, with optional overrides
        ``label`` and the ``"label"`` entry in the ``series_params`` option (see
        :attr:`options`). ``label`` is returned if it is not ``None``, as this is the
        override with the highest priority. If it is ``None``, then the drawer will look
        for a ``"label"`` entry in ``series_params``, for the series identified by
        ``name``. If this entry doesn't exist, or is ``None``, then ``name`` is used as
        the label. If all these options are ``None``, then ``None`` is returned;
        signifying that a legend entry for the provided series should not be generated.
        Note that :meth:`label_for` will convert ``name`` to ``str`` when it is
        returned.

        Args:
            name: The name of the series.
            label: Optional label override.

        Returns:
            The legend entry label, or ``None``.
        """
        if label is not None:
            return str(label)

        if name is not None:
            return self.figure_options.series_params.get(name, {}).get("label", str(name))
        return None

    @abstractmethod
    def scatter(
        self,
        x_data: Sequence[float],
        y_data: Sequence[float],
        x_err: Optional[Sequence[float]] = None,
        y_err: Optional[Sequence[float]] = None,
        name: Optional[SeriesName] = None,
        label: Optional[str] = None,
        legend: bool = False,
        **options,
    ):
        """Draw scatter points, with optional error-bars.

        Args:
            x_data: X values.
            y_data: Y values.
            x_err: Optional error for X values.
            y_err: Optional error for Y values.
            name: Name of this series.
            label: Optional legend label to override ``name`` and ``series_params``.
            legend: Whether the drawn area must have a legend entry. Defaults to False.
                The series label in the legend will be ``label`` if it is not None. If
                it is, then ``series_params`` is searched for a ``"label"`` entry for
                the series identified by ``name``. If this is also ``None``, then
                ``name`` is used as the fallback. If no ``name`` is provided, then no
                legend entry is generated.
            options: Valid options for the drawer backend API.
        """

    @abstractmethod
    def line(
        self,
        x_data: Sequence[float],
        y_data: Sequence[float],
        name: Optional[SeriesName] = None,
        label: Optional[str] = None,
        legend: bool = False,
        **options,
    ):
        """Draw a line.

        Args:
            x_data: X values.
            y_data: Y values.
            name: Name of this series.
            label: Optional legend label to override ``name`` and ``series_params``.
            legend: Whether the drawn area must have a legend entry. Defaults to False.
                The series label in the legend will be ``label`` if it is not None. If
                it is, then ``series_params`` is searched for a ``"label"`` entry for
                the series identified by ``name``. If this is also ``None``, then
                ``name`` is used as the fallback. If no ``name`` is provided, then no
                legend entry is generated.
            options: Valid options for the drawer backend API.
        """

    @abstractmethod
    def hline(
        self,
        y_value: float,
        name: Optional[SeriesName] = None,
        label: Optional[str] = None,
        legend: bool = False,
        **options,
    ):
        """Draw a horizontal line.

        Args:
            y_value: Y value for line.
            name: Name of this series.
            label: Optional legend label to override ``name`` and ``series_params``.
            legend: Whether the drawn area must have a legend entry. Defaults to False.
                The series label in the legend will be ``label`` if it is not None. If
                it is, then ``series_params`` is searched for a ``"label"`` entry for
                the series identified by ``name``. If this is also ``None``, then
                ``name`` is used as the fallback. If no ``name`` is provided, then no
                legend entry is generated.
            options: Valid options for the drawer backend API.
        """

    @abstractmethod
    def filled_y_area(
        self,
        x_data: Sequence[float],
        y_ub: Sequence[float],
        y_lb: Sequence[float],
        name: Optional[SeriesName] = None,
        label: Optional[str] = None,
        legend: bool = False,
        **options,
    ):
        """Draw filled area as a function of x-values.

        Args:
            x_data: X values.
            y_ub: The upper boundary of Y values.
            y_lb: The lower boundary of Y values.
            name: Name of this series.
            label: Optional legend label to override ``name`` and ``series_params``.
            legend: Whether the drawn area must have a legend entry. Defaults to False.
                The series label in the legend will be ``label`` if it is not None. If
                it is, then ``series_params`` is searched for a ``"label"`` entry for
                the series identified by ``name``. If this is also ``None``, then
                ``name`` is used as the fallback. If no ``name`` is provided, then no
                legend entry is generated.
            options: Valid options for the drawer backend API.
        """

    @abstractmethod
    def filled_x_area(
        self,
        x_ub: Sequence[float],
        x_lb: Sequence[float],
        y_data: Sequence[float],
        name: Optional[SeriesName] = None,
        label: Optional[str] = None,
        legend: bool = False,
        **options,
    ):
        """Draw filled area as a function of y-values.

        Args:
            x_ub: The upper boundary of X values.
            x_lb: The lower boundary of X values.
            y_data: Y values.
            name: Name of this series.
            label: Optional legend label to override ``name`` and ``series_params``.
            legend: Whether the drawn area must have a legend entry. Defaults to False.
                The series label in the legend will be ``label`` if it is not None. If
                it is, then ``series_params`` is searched for a ``"label"`` entry for
                the series identified by ``name``. If this is also ``None``, then
                ``name`` is used as the fallback. If no ``name`` is provided, then no
                legend entry is generated.
            options: Valid options for the drawer backend API.
        """

    @abstractmethod
    def textbox(
        self,
        description: str,
        rel_pos: Optional[Tuple[float, float]] = None,
        **options,
    ):
        """Draw text box.

        Args:
            description: A string to be drawn inside a report box.
            rel_pos: Relative position of the text-box. If None, the default
                ``textbox_rel_pos`` from the style is used.
            options: Valid options for the drawer backend API.
        """

    @abstractmethod
    def image(
        self,
        data: np.ndarray,
        extent: Optional[ExtentTuple] = None,
        name: Optional[SeriesName] = None,
        label: Optional[str] = None,
        cmap: Optional[Union[str, Any]] = None,
        cmap_use_series_colors: bool = False,
        colorbar: bool = False,
        **options,
    ):
        """Draw an image of numerical values, series names, or RGB/A values.

        Args:
            data: The two-/three-dimensional data defining an image. If
                ``data.dims==2``, then the pixel colors are determined by ``cmap`` and
                ``cmap_use_series_colors``. If ``data.dims==3``, then it is assumed that
                ``data`` contains either RGB or RGBA data; which requires the third
                dimension to have length ``3`` or ``4`` respectively. For RGB/A data,
                the elements of ``data`` must be floats or integers in the range 0-1 and
                0-255 respectively. If the data is two-dimensional, there is no limit on
                the range of the values if they are numerical. If
                ``cmap_use_series_colors=True``, then ``data`` contains series names;
                which can be strings or numerical values, as long as they are
                appropriate series names.
            extent: An optional tuple ``(x_min, x_max, y_min, y_max)`` which defines a
                rectangular region within which the values inside ``data`` should be
                plotted. The units of ``extent`` are the same as those of the X and Y
                axes for the axis. If None, the extent of the image is taken as ``(0,
                data.shape[0], 0, data.shape[1])``. Default is None.
            name: Name of this image. Used to lookup ``canvas`` and ``label`` in
                ``series_params``.
            label: An optional label for the colorbar, if ``colorbar=True``.
            cmap: Optional colormap for assigning colors to the image values, if
                ``data`` is not an RGB/A image. ``cmap`` must be a string or object
                instance which is recognized by the drawer. Defaults to None.
            cmap_use_series_colors: Whether to assign colors to the image based on
                series colors, where the values inside ``data`` are series names. If
                ``cmap_use_series_colors=True``,``cmap`` is ignored. This only works for
                two-dimensional images as three-dimensional ``data`` contains explicit
                colors as RGB/A values. If ``len(data.shape)=3``,
                ``cmap_use_series_colours`` is ignored. Defaults to False.
            colorbar: Whether to add a bar showing the color-value mapping for the
                image. Defaults to False.
            options: Valid options for the drawer backend API.
        """

    @property
    @abstractmethod
    def figure(self):
        """Return figure object handler to be saved in the database."""

    def config(self) -> Dict:
        """Return the config dictionary for this drawer."""
        options = dict((key, getattr(self._options, key)) for key in self._set_options)
        figure_options = dict(
            (key, getattr(self._figure_options, key)) for key in self._set_figure_options
        )

        return {
            "cls": type(self),
            "options": options,
            "figure_options": figure_options,
        }

    def __json_encode__(self):
        return self.config()

    @classmethod
    def __json_decode__(cls, value):
        instance = cls()
        if "options" in value:
            instance.set_options(**value["options"])
        if "figure_options" in value:
            instance.set_figure_options(**value["figure_options"])
        return instance



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
A library of fit functions.
"""
# pylint: disable=invalid-name, no-member
import numpy as np


def cos(
    x: np.ndarray,
    amp: float = 1.0,
    freq: float = 1 / (2 * np.pi),
    phase: float = 0.0,
    baseline: float = 0.0,
) -> np.ndarray:
    r"""Cosine function.

    .. math::
        y = {\rm amp} \cdot \cos\left(2 \pi {\rm freq} \cdot x
            + {\rm phase}\right) + {\rm baseline}
    """
    return amp * np.cos(2 * np.pi * freq * x + phase) + baseline


def sin(
    x: np.ndarray,
    amp: float = 1.0,
    freq: float = 1 / (2 * np.pi),
    phase: float = 0.0,
    baseline: float = 0.0,
) -> np.ndarray:
    r"""Sine function.

    .. math::
        y = {\rm amp} \cdot \sin\left(2 \pi {\rm freq} \cdot x
            + {\rm phase}\right) + {\rm baseline}
    """
    return amp * np.sin(2 * np.pi * freq * x + phase) + baseline


def exponential_decay(
    x: np.ndarray,
    amp: float = 1.0,
    lamb: float = 1.0,
    base: float = np.e,
    x0: float = 0.0,
    baseline: float = 0.0,
) -> np.ndarray:
    r"""Exponential function

    .. math::
        y = {\rm amp} \cdot {\rm base}^{\left( - \lambda x + {\rm x0} \right)} + {\rm baseline}
    """
    return amp * base ** (-lamb * x + x0) + baseline


def gaussian(
    x: np.ndarray, amp: float = 1.0, sigma: float = 1.0, x0: float = 0.0, baseline: float = 0.0
) -> np.ndarray:
    r"""Gaussian function

    .. math::
        y = {\rm amp} \cdot \exp \left( - (x - x0)^2 / 2 \sigma^2 \right) + {\rm baseline}
    """
    return amp * np.exp(-((x - x0) ** 2) / (2 * sigma**2)) + baseline


def sqrt_lorentzian(
    x: np.ndarray, amp: float = 1.0, kappa: float = 1.0, x0: float = 0.0, baseline: float = 0.0
) -> np.ndarray:
    r"""Square-root Lorentzian function for spectroscopy.

    .. math::
        y = \frac{{\rm amp} |\kappa|}{\sqrt{\kappa^2 + 4(x -x_0)^2}} + {\rm baseline}
    """
    return amp * np.abs(kappa) / np.sqrt(kappa**2 + 4 * (x - x0) ** 2) + baseline


def cos_decay(
    x: np.ndarray,
    amp: float = 1.0,
    tau: float = 1.0,
    freq: float = 1 / (2 * np.pi),
    phase: float = 0.0,
    baseline: float = 0.0,
) -> np.ndarray:
    r"""Cosine function with exponential decay.

    .. math::
        y = {\rm amp} \cdot e^{-x/\tau} \cos\left(2 \pi \cdot {\rm freq} \cdot x
        + {\rm phase}\right) + {\rm baseline}
    """
    return exponential_decay(x, lamb=1 / tau) * cos(x, amp=amp, freq=freq, phase=phase) + baseline


def sin_decay(
    x: np.ndarray,
    amp: float = 1.0,
    tau: float = 1.0,
    freq: float = 1 / (2 * np.pi),
    phase: float = 0.0,
    baseline: float = 0.0,
) -> np.ndarray:
    r"""Sine function with exponential decay.

    .. math::
        y = {\rm amp} \cdot e^{-x/\tau} \sin\left(2 \pi \cdot {\rm freq} \cdot x
        + {\rm phase}\right) + {\rm baseline}
    """
    return exponential_decay(x, lamb=1 / tau) * sin(x, amp=amp, freq=freq, phase=phase) + baseline



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Utils in curve analysis."""

from typing import Union, Optional, List, Dict, Tuple, Callable
import time

import asteval
import lmfit
import numpy as np
import pandas as pd
from qiskit.utils.deprecation import deprecate_func
from qiskit.utils import detach_prefix
from uncertainties import UFloat, wrap as wrap_function
from uncertainties import unumpy

from qiskit_experiments.curve_analysis.curve_data import CurveFitResult
from qiskit_experiments.exceptions import AnalysisError, QiskitError
from qiskit_experiments.framework import AnalysisResultData


UNUMPY_FUNCS = {fn: getattr(unumpy, fn) for fn in unumpy.__all__}


def is_error_not_significant(
    val: Union[float, UFloat],
    fraction: float = 1.0,
    absolute: Optional[float] = None,
) -> bool:
    """Check if the standard error of given value is not significant.

    Args:
        val: Input value to evaluate. This is assumed to be float or ufloat.
        fraction: Valid fraction of the nominal part to its standard error.
            This function returns ``False`` if the nominal part is
            smaller than the error by this fraction.
        absolute: Use this value as a threshold if given.

    Returns:
        ``True`` if the standard error of given value is not significant.
    """
    if isinstance(val, float):
        return True

    threshold = absolute if absolute is not None else fraction * val.nominal_value
    if np.isnan(val.std_dev) or val.std_dev < threshold:
        return True

    return False


def analysis_result_to_repr(result: AnalysisResultData) -> str:
    """A helper function to create string representation from analysis result data object.

    Args:
        result: Analysis result data.

    Returns:
        String representation of the data.

    Raises:
        AnalysisError: When the result data is not likely fit parameter.
    """
    if not isinstance(result.value, (float, UFloat)):
        raise AnalysisError(f"Result data {result.name} is not a valid fit parameter data type.")

    unit = result.extra.get("unit", None)

    def _format_val(value):
        # Return value with unit with prefix, i.e. 1000 Hz -> 1 kHz.
        if unit:
            try:
                val, val_prefix = detach_prefix(value, decimal=3)
            except ValueError:
                val = value
                val_prefix = ""
            return f"{val: .3g}", f" {val_prefix}{unit}"
        if np.abs(value) < 1e-3 or np.abs(value) > 1e3:
            return f"{value: .4e}", ""
        return f"{value: .4g}", ""

    if isinstance(result.value, float):
        # Only nominal part
        n_repr, n_unit = _format_val(result.value)
        value_repr = n_repr + n_unit
    else:
        # Nominal part
        n_repr, n_unit = _format_val(result.value.nominal_value)

        # Standard error part
        if result.value.std_dev is not None and np.isfinite(result.value.std_dev):
            s_repr, s_unit = _format_val(result.value.std_dev)
            if n_unit == s_unit:
                value_repr = f" {n_repr} \u00B1 {s_repr}{n_unit}"
            else:
                value_repr = f" {n_repr + n_unit} \u00B1 {s_repr + s_unit}"
        else:
            value_repr = n_repr + n_unit

    return f"{result.name} = {value_repr}"


def convert_lmfit_result(
    result: lmfit.minimizer.MinimizerResult,
    models: List[lmfit.Model],
    xdata: np.ndarray,
    ydata: np.ndarray,
    residuals: np.ndarray,
) -> CurveFitResult:
    """A helper function to convert LMFIT ``MinimizerResult`` into :class:`.CurveFitResult`.

    :class:`.CurveFitResult` is a dataclass that can be serialized with the experiment JSON decoder.
    In addition, this class converts LMFIT ``Parameter`` objects into ufloats so that
    extra parameter computation in the analysis post-processing can perform
    accurate error propagation with parameter correlation.

    Args:
        result: Output from LMFIT ``minimize``.
        models: Model used for the fitting. Function description is extracted.
        xdata: X values used for the fitting.
        ydata: Y values used for the fitting.
        residuals: The residuals of the ydata from the model.

    Returns:
        QiskitExperiments :class:`.CurveFitResult` object.
    """
    model_descriptions = {}
    for model in models:
        if hasattr(model, "expr"):
            func_repr = model.expr
        else:
            signature = ", ".join(model.independent_vars + model.param_names)
            func_repr = f"F({signature})"
        model_descriptions[model._name] = func_repr

    if result is None:
        return CurveFitResult(
            model_repr=model_descriptions,
            success=False,
            x_data=xdata,
            y_data=ydata,
        )

    covar = getattr(result, "covar", None)
    if covar is not None and np.any(np.diag(covar) < 0):
        covar = None

    return CurveFitResult(
        method=result.method,
        model_repr=model_descriptions,
        success=result.success,
        nfev=result.nfev,
        message=result.message,
        dof=result.nfree,
        init_params={name: param.init_value for name, param in result.params.items()},
        chisq=result.chisqr,
        reduced_chisq=result.redchi,
        aic=result.aic,
        bic=result.bic,
        params={name: param.value for name, param in result.params.items()},
        var_names=result.var_names,
        x_data=xdata,
        y_data=ydata,
        weighted_residuals=result.residual,
        residuals=residuals,
        covar=covar,
    )


def eval_with_uncertainties(
    x: np.ndarray,
    model: lmfit.Model,
    params: Dict[str, UFloat],
) -> np.ndarray:
    """Compute Y values with error propagation.

    Args:
        x: X values.
        model: LMFIT model.
        params: Fitter parameters in correlated ufloats.

    Returns:
        Y values with uncertainty (uarray).
    """
    sub_params = {name: params[name] for name in model.param_names}

    if hasattr(model, "expr"):
        # If the model has string expression, we regenerate unumpy fit function.
        # Note that propagating the error through the function requires computation of
        # derivatives, which is done by uncertainties.wrap (or perhaps manually).
        # However, usually computation of derivative is heavy computing overhead,
        # and it is better to use hard-coded derivative functions if it is known.
        # The unumpy functions provide such derivatives, and it's much faster.
        # Here we parse the expression with ASTEVAL, and replace the mapping to
        # the functions in Python's math or numpy with one in unumpy module.
        # Benchmarking with RamseyXY experiment with 100 data points,
        # this yields roughly 60% computation time reduction.
        interpreter = asteval.Interpreter()
        astcode = interpreter.parse(model.expr.strip())

        # Replace function with unumpy version
        interpreter.symtable.update(UNUMPY_FUNCS)
        # Add parameters
        interpreter.symtable.update(sub_params)
        # Add x values
        interpreter.symtable["x"] = x

        interpreter.start_time = time.time()
        try:
            return interpreter.run(astcode)
        except Exception:  # pylint: disable=broad-except
            # User provided function does not support ufloats.
            # Likely using not defined function in unumpy.
            # This falls into normal derivative computation.
            pass

    wrapfunc = np.vectorize(wrap_function(model.func))
    return wrapfunc(x=x, **sub_params)


def shot_weighted_average(
    yvals: np.ndarray,
    yerrs: np.ndarray,
    shots: np.ndarray,
) -> Tuple[float, float, float]:
    """Compute shot based variance and weighted average of the categorized data frame.

    Sample is weighted by the shot number.

    Args:
        yvals: Y values to average.
        yerrs: Y errors to average.
        shots: Number of shots used to obtain Y value and error.

    Returns:
        Averaged Y value, Y error, and total shots.
    """
    if len(yvals) == 1:
        return yvals[0], yerrs[0], shots[0]

    if any(s is pd.NA for s in shots):
        # Shot number is unknown
        return np.mean(yvals), np.nan, pd.NA

    total_shots = np.sum(shots)
    weights = shots / total_shots

    avg_yval = np.sum(weights * yvals)
    avg_yerr = np.sqrt(np.sum(weights**2 * yerrs**2))

    return avg_yval, avg_yerr, total_shots


def inverse_weighted_variance(
    yvals: np.ndarray,
    yerrs: np.ndarray,
    shots: np.ndarray,
) -> Tuple[float, float, int]:
    """Compute inverse weighted variance and weighted average of the categorized data frame.

    Sample is weighted by the inverse of the data variance.

    Args:
        yvals: Y values to average.
        yerrs: Y errors to average.
        shots: Number of shots used to obtain Y value and error.

    Returns:
        Averaged Y value, Y error, and total shots.
    """
    if len(yvals) == 1:
        return yvals[0], yerrs[0], shots[0]

    total_shots = np.sum(shots)
    weights = 1 / yerrs**2
    yvar = 1 / np.sum(weights)

    avg_yval = yvar * np.sum(weights * yvals)
    avg_yerr = np.sqrt(yvar)

    return avg_yval, avg_yerr, total_shots


# pylint: disable=unused-argument
def sample_average(
    yvals: np.ndarray,
    yerrs: np.ndarray,
    shots: np.ndarray,
) -> Tuple[float, float, int]:
    """Compute sample based variance and average of the categorized data frame.

    Original variance of the data is ignored and variance is computed with the y values.

    Args:
        yvals: Y values to average.
        yerrs: Y errors to average (ignored).
        shots: Number of shots used to obtain Y value and error.

    Returns:
        Averaged Y value, Y error, and total shots.
    """
    if len(yvals) == 1:
        return yvals[0], 0.0, shots[0]

    total_shots = np.sum(shots)

    avg_yval = np.mean(yvals)
    avg_yerr = np.sqrt(np.mean((avg_yval - yvals) ** 2) / len(yvals))

    return avg_yval, avg_yerr, total_shots


@deprecate_func(
    since="0.6",
    additional_msg="The curve data representation has been replaced by the `DataFrame` format.",
    package_name="qiskit-experiments",
    pending=True,
)
def filter_data(data: List[Dict[str, any]], **filters) -> List[Dict[str, any]]:
    """Return the list of filtered data

    Args:
        data: list of data dicts.
        filters: kwargs for filtering based on metadata
                 values.

    Returns:
        The list of filtered data. If no filters are provided this will be the
        input list.
    """
    if not filters:
        return data
    filtered_data = []
    for datum in data:
        include = True
        metadata = datum["metadata"]
        for key, val in filters.items():
            if key not in metadata or metadata[key] != val:
                include = False
                break
        if include:
            filtered_data.append(datum)
    return filtered_data


@deprecate_func(
    since="0.6",
    additional_msg="The curve data representation has been replaced by the `DataFrame` format.",
    package_name="qiskit-experiments",
    pending=True,
)
def mean_xy_data(
    xdata: np.ndarray,
    ydata: np.ndarray,
    sigma: Optional[np.ndarray] = None,
    shots: Optional[np.ndarray] = None,
    method: str = "sample",
) -> Tuple[np.ndarray, ...]:
    r"""Return (x, y_mean, sigma) data.

    The mean is taken over all :math:`y` data values with the same :math:`x` data value using
    the specified method. For each :math:`x` the mean :math:`\overline{y}` and variance
    :math:`\sigma^2` are computed as

    * ``"sample"`` (default): *Sample mean and variance*

      * :math:`\overline{y} = \sum_{i=1}^N y_i / N`,

      * :math:`\sigma^2 = \sum_{i=1}^N ((\overline{y} - y_i)^2) / N`

    * ``"iwv"``: *Inverse-weighted variance*

      * :math:`\overline{y} = (\sum_{i=1}^N y_i / \sigma_i^2 ) \sigma^2`
      * :math:`\sigma^2 = 1 / (\sum_{i=1}^N 1 / \sigma_i^2)`

    * ``"shots_weighted_variance"``: *Sample mean and variance with weights from shots*

      * :math:`\overline{y} = \sum_{i=1}^N n_i y_i / M`,

      * :math:`\sigma^2 = \sum_{i=1}^N (n_i \sigma_i / M)^2`,
        where :math:`n_i` is the number of shots per data point and :math:`M = \sum_{i=1}^N n_i`
        is a total number of shots from different circuit execution at the same :math:`x` value.
        If ``shots`` is not provided, this applies uniform weights to all values.

    Args:
        xdata: 1D or 2D array of xdata from curve_fit_data or
            multi_curve_fit_data
        ydata: array of ydata returned from curve_fit_data or
            multi_curve_fit_data
        sigma: Optional, array of standard deviations in ydata.
        shots: Optional, array of shots used to get a data point.
        method: The method to use for computing y means and
            standard deviations sigma (default: "sample").

    Returns:
        tuple: ``(x, y_mean, sigma, shots)``, where ``x`` is an arrays of unique
        x-values, ``y`` is an array of sample mean y-values, ``sigma`` is an array of
        sample standard deviation of y values, and ``shots`` are the total number of
        experiment shots used to evaluate the data point. If ``shots`` in the function
        call is ``None``, the numbers appear in the returned value will represent just a
        number of duplicated x value entries.

    Raises:
        QiskitError: If the "ivw" method is used without providing a sigma.
    """
    x_means = np.unique(xdata, axis=0)
    y_means = np.zeros(x_means.size)
    y_sigmas = np.zeros(x_means.size)
    y_shots = np.zeros(x_means.size)

    if shots is None or any(np.isnan(shots)):
        # this will become standard average
        shots = np.ones_like(xdata)

    # Sample mean and variance method
    if method == "sample":
        for i in range(x_means.size):
            # Get positions of y to average
            idxs = xdata == x_means[i]
            ys = ydata[idxs]
            ns = shots[idxs]

            # Compute sample mean and sample standard error of the mean
            y_means[i] = np.mean(ys)
            y_sigmas[i] = np.sqrt(np.mean((y_means[i] - ys) ** 2) / ys.size)
            y_shots[i] = np.sum(ns)

        return x_means, y_means, y_sigmas, y_shots

    # Inverse-weighted variance method
    if method == "iwv":
        if sigma is None:
            raise QiskitError(
                "The inverse-weighted variance method cannot be used with `sigma=None`"
            )
        for i in range(x_means.size):
            # Get positions of y to average
            idxs = xdata == x_means[i]
            ys = ydata[idxs]
            ns = shots[idxs]

            # Compute the inverse-variance weighted y mean and variance
            weights = 1 / sigma[idxs] ** 2
            y_var = 1 / np.sum(weights)
            y_means[i] = y_var * np.sum(weights * ys)
            y_sigmas[i] = np.sqrt(y_var)
            y_shots[i] = np.sum(ns)

        return x_means, y_means, y_sigmas, y_shots

    # Quadrature sum of variance
    if method == "shots_weighted":
        for i in range(x_means.size):
            # Get positions of y to average
            idxs = xdata == x_means[i]
            ys = ydata[idxs]
            ss = sigma[idxs]
            ns = shots[idxs]
            weights = ns / np.sum(ns)

            # Compute sample mean and sum of variance with weights based on shots
            y_means[i] = np.sum(weights * ys)
            y_sigmas[i] = np.sqrt(np.sum(weights**2 * ss**2))
            y_shots[i] = np.sum(ns)

        return x_means, y_means, y_sigmas, y_shots

    # Invalid method
    raise QiskitError(f"Unsupported method {method}")


@deprecate_func(
    since="0.6",
    additional_msg="The curve data representation has been replaced by the `DataFrame` format.",
    package_name="qiskit-experiments",
    pending=True,
)
def multi_mean_xy_data(
    series: np.ndarray,
    xdata: np.ndarray,
    ydata: np.ndarray,
    sigma: Optional[np.ndarray] = None,
    shots: Optional[np.ndarray] = None,
    method: str = "sample",
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Take mean of multi series data set. See :func:`.mean_xy_data`.

    Args:
        series: Series index.
        xdata: 1D or 2D array of xdata from curve_fit_data or
               multi_curve_fit_data
        ydata: array of ydata returned from curve_fit_data or
               multi_curve_fit_data
        sigma: Optional, array of standard deviations in ydata.
        shots: Optional, array of shots used to get a data point.
        method: The method to use for computing y means and
                standard deviations sigma (default: "sample").

    Returns:
        Tuple of ``(series, xdata, ydata, sigma, shots)``.

    """
    series_vals = np.unique(series)

    series_means = []
    xdata_means = []
    ydata_means = []
    sigma_means = []
    shots_sums = []

    # Get x, y, sigma data for series and process mean data
    for series_val in series_vals:
        idxs = series == series_val
        sigma_i = sigma[idxs] if sigma is not None else None
        shots_i = shots[idxs] if shots is not None else None

        x_mean, y_mean, sigma_mean, shots_sum = mean_xy_data(
            xdata[idxs], ydata[idxs], sigma=sigma_i, shots=shots_i, method=method
        )
        series_means.append(np.full(x_mean.size, series_val, dtype=int))
        xdata_means.append(x_mean)
        ydata_means.append(y_mean)
        sigma_means.append(sigma_mean)
        shots_sums.append(shots_sum)

    # Concatenate lists
    return (
        np.concatenate(series_means),
        np.concatenate(xdata_means),
        np.concatenate(ydata_means),
        np.concatenate(sigma_means),
        np.concatenate(shots_sums),
    )


@deprecate_func(
    since="0.6",
    additional_msg="The curve data representation has been replaced by the `DataFrame` format.",
    package_name="qiskit-experiments",
    pending=True,
)
def data_sort(
    series: np.ndarray,
    xdata: np.ndarray,
    ydata: np.ndarray,
    sigma: Optional[np.ndarray] = None,
    shots: Optional[np.ndarray] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Sort data.

    Input x values may not be lined up in order, since experiment may accept user input array,
    or data may be concatenated with previous scan. This sometimes confuses the algorithmic
    generation of initial guesses especially when guess depends on derivative.

    This returns data set that is sorted by xdata and series in ascending order.

    Args:
        series: Series index.
        xdata: 1D or 2D array of xdata.
        ydata: Array of ydata.
        sigma: Optional, array of standard deviations in ydata.
        shots: Optional, array of shots used to get a data point.

    Returns:
        Tuple of (series, xdata, ydata, sigma, shots) sorted in ascending order of xdata
        and series.
    """
    if sigma is None:
        sigma = np.full(series.size, np.nan, dtype=float)

    if shots is None:
        shots = np.full(series.size, np.nan, dtype=float)

    sorted_data = sorted(zip(series, xdata, ydata, sigma, shots), key=lambda d: (d[0], d[1]))

    return np.asarray(sorted_data).T


def level2_probability(data: Dict[str, any], outcome: str) -> Tuple[float, float]:
    """Return the outcome probability mean and variance.

    Args:
        data: A data dict containing count data.
        outcome: bitstring for desired outcome probability.

    Returns:
        tuple: (p_mean, p_var) of the probability mean and variance
               estimated from the counts.

    .. note::

        This assumes a binomial distribution where :math:`K` counts
        of the desired outcome from :math:`N` shots the
        mean probability is :math:`p = K / N` and the variance is
        :math:`\\sigma^2 = p (1-p) / N`.
    """
    counts = data["counts"]

    shots = sum(counts.values())
    p_mean = counts.get(outcome, 0.0) / shots
    p_var = p_mean * (1 - p_mean) / shots
    return p_mean, p_var


def probability(outcome: str) -> Callable:
    """Return probability data processor callback used by the analysis classes."""

    def data_processor(data):
        return level2_probability(data, outcome)

    return data_processor



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Configurable stylesheet for :class:`BasePlotter` and :class:`BaseDrawer`.
"""


class PlotStyle(dict):
    """A stylesheet for :class:`BasePlotter` and :class:`BaseDrawer`.

    This style class is used by :class:`BasePlotter` and :class:`BaseDrawer`. The
    default style for Qiskit Experiments is defined in :meth:`default_style`. Style
    parameters are stored as dictionary entries, grouped by graphics or figure
    component. For example, style parameters relating to textboxes have the prefix
    ``textbox_``. For default style parameter names and their values, see the
    :meth:`default_style` method.

    .. rubric:: Example
    .. code-block:: python

        # Create custom style
        custom_style = PlotStyle(
            {
                "legend_loc": "upper right",
                "textbox_rel_pos": (1, 1),
                "textbox_text_size": 14,
            }
        )

        # Create full style instance by combining with default style.
        full_style = PlotStyle.merge(PlotStyle.default_style(), custom_style)

        ## Query style parameters
        # Returns "upper right"
        full_style["legend_loc"]
        # Returns the value provided in ``PlotStyle.default_style()``
        full_style["axis_label_size"]
    """

    @classmethod
    def default_style(cls) -> "PlotStyle":
        # Ignore pylint warnings as `Style Parameters` doesn't contain method parameters.
        # pylint: disable=differing-param-doc,differing-type-doc
        """The default style across Qiskit Experiments.

        The following is a description of the default style parameters are what they are
        used for.

        Style Parameters:
            figsize (Tuple[int,int]): The size of the figure ``(width, height)``, in
                inches.
            legend_loc (Optional[str]): The location of the legend in axis coordinates.
                If None, location is automatically determined by the drawer.
            tick_label_size (int): The font size for tick labels.
            axis_label_size (int): The font size for axis labels.
            symbol_size (float): The size of symbols for points/markers, proportional to
                the area of the drawn graphic.
            errorbar_capsize (float): The size of end-caps for error-bars.
            textbox_rel_pos (Tuple[float,float]): The relative position ``(horizontal,
                vertical)`` of textboxes, as a percentage of the canvas dimensions.
            textbox_text_size (int): The font size for textboxes.

        Returns:
            The default plot style used by Qiskit Experiments.
        """
        style = {
            # size of figure (width, height)
            "figsize": (8, 5),  # Tuple[int, int]
            # legend location (vertical, horizontal) or None.
            "legend_loc": None,  # str
            # size of tick label
            "tick_label_size": 14,  #  int
            # size of axis label
            "axis_label_size": 16,  # int
            # relative position of a textbox
            "textbox_rel_pos": (0.5, -0.25),  # Tuple[float, float]
            # size of textbox text
            "textbox_text_size": 12,  # int
            # size of caps for error-bars
            "errorbar_capsize": 4,  # float
            # Default size of symbols, used for graphics where symbols are drawn for points.
            "symbol_size": 6.0,  # float
        }
        return cls(**style)

    @classmethod
    def merge(cls, style1: "PlotStyle", style2: "PlotStyle") -> "PlotStyle":
        """Merge ``style2`` into ``style1`` as a new PlotStyle instance.

        This method merges an additional style ``style2`` into a base instance
        ``style1``, returning the merged style instance instead of modifying the inputs.

        Args:
            style1: Base PlotStyle instance.
            style2: Additional PlotStyle instance.

        Returns:
            Merged style instance.
        """
        return PlotStyle({**style1, **style2})



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Fine frequency calibration experiment."""

from typing import Dict, List, Optional, Sequence
import numpy as np

from qiskit.providers.backend import Backend
from qiskit.circuit import QuantumCircuit

from qiskit_experiments.framework import ExperimentData
from qiskit_experiments.calibration_management.update_library import BaseUpdater
from qiskit_experiments.calibration_management import (
    BaseCalibrationExperiment,
    Calibrations,
)
from qiskit_experiments.library.characterization.fine_frequency import FineFrequency


class FineFrequencyCal(BaseCalibrationExperiment, FineFrequency):
    """A calibration version of the fine frequency experiment."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        backend: Optional[Backend] = None,
        cal_parameter_name: Optional[str] = "drive_freq",
        delay_duration: Optional[int] = None,
        repetitions: List[int] = None,
        auto_update: bool = True,
        gate_name: str = "sx",
    ):
        r"""See class :class:`.FineFrequency` for details.

        Note that this class implicitly assumes that the target angle of the gate
        is :math:`\pi/2` as seen from the default analysis options. This experiment
        can be seen as a calibration of a finite duration ``rz(pi/2)`` gate with any
        error attributed to a frequency offset in the qubit.

        Args:
            physical_qubits: Sequence containing the qubit for which to run the
                fine frequency calibration.
            calibrations: The calibrations instance with the schedules.
            backend: Optional, the backend to run the experiment on.
            cal_parameter_name: The name of the parameter to update in the calibrations.
                This defaults to `drive_freq`.
            delay_duration: The duration of the delay at :math:`n=1`. If this value is
                not given then the duration of the gate named ``gate_name`` in the
                calibrations will be used.
            auto_update: Whether to automatically update the calibrations or not. By
                default, this variable is set to True.
            gate_name: This argument is only needed if ``delay_duration`` is None. This
                should be the name of a valid schedule in the calibrations.
        """
        if delay_duration is None:
            delay_duration = calibrations.get_schedule(gate_name, physical_qubits[0]).duration

        super().__init__(
            calibrations,
            physical_qubits,
            delay_duration=delay_duration,
            schedule_name=None,
            repetitions=repetitions,
            backend=backend,
            cal_parameter_name=cal_parameter_name,
            auto_update=auto_update,
        )

        if self.backend is not None:
            self.set_experiment_options(dt=self._backend_data.dt)

    @classmethod
    def _default_experiment_options(cls):
        """default values for the fine frequency calibration experiment.

        Experiment Options:
            dt (float): The duration of the time unit ``dt`` of the delay and schedules in seconds.
        """
        options = super()._default_experiment_options()
        options.dt = None
        return options

    def _metadata(self) -> Dict[str, any]:
        """Add metadata to the experiment data making it more self contained.

        The following keys are added to the experiment's metadata:
            cal_param_value: The value of the drive frequency parameter. This value together with
                the fit result will be used to find the new value of the drive frequency parameter.
            cal_param_name: The name of the parameter in the calibrations.
            cal_group: The calibration group to which the parameter belongs.
            delay_duration: The duration of the first delay.
            dt: The number of ``dt`` units of the delay.
        """
        metadata = super()._metadata()
        metadata["delay_duration"] = self.experiment_options.delay_duration
        metadata["dt"] = self.experiment_options.dt
        metadata["cal_param_value"] = self._cals.get_parameter_value(
            self._param_name,
            self.physical_qubits,
            group=self.experiment_options.group,
        )

        return metadata

    def _attach_calibrations(self, circuit: QuantumCircuit):
        """Adds the calibrations to the transpiled circuits."""
        schedule = self._cals.get_schedule("sx", self.physical_qubits)
        circuit.add_calibration("sx", self.physical_qubits, schedule)

    def update_calibrations(self, experiment_data: ExperimentData):
        r"""Update the qubit frequency based on the measured angle deviation.

        The frequency of the qubit is updated according to

        .. math::

            f \to f - \frac{{\rm d}\theta}{2\pi\tau{\rm d}t}

        Here, :math:`{\rm d}\theta` is the measured angle error from the fit. The duration of
        the single qubit-gate is :math:`\tau` in samples and :math:`{\rm d}t` is the duration
        of a sample. This is also the duration of the time unit ``dt`` of the delay.
        """
        result_index = self.experiment_options.result_index
        group = experiment_data.metadata["cal_group"]
        prev_freq = experiment_data.metadata["cal_param_value"]
        tau = experiment_data.metadata["delay_duration"]
        dt = experiment_data.metadata["dt"]

        d_theta = BaseUpdater.get_value(experiment_data, "d_theta", result_index)
        new_freq = prev_freq + d_theta / (2 * np.pi * tau * dt)

        BaseUpdater.add_parameter_value(
            self._cals, experiment_data, new_freq, self._param_name, self._sched_name, group
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Calibration version of spectroscopy experiments."""

from typing import Iterable, Optional, Sequence

from qiskit.circuit import QuantumCircuit
from qiskit.providers.backend import Backend

from qiskit_experiments.library.characterization.qubit_spectroscopy import QubitSpectroscopy
from qiskit_experiments.library.characterization.ef_spectroscopy import EFSpectroscopy
from qiskit_experiments.calibration_management.update_library import Frequency
from qiskit_experiments.calibration_management.calibrations import Calibrations
from qiskit_experiments.calibration_management.base_calibration_experiment import (
    BaseCalibrationExperiment,
)


class RoughFrequencyCal(BaseCalibrationExperiment, QubitSpectroscopy):
    """A calibration experiment that runs :class:`.QubitSpectroscopy` to calibrate the qubit
    transition frequency."""

    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        frequencies: Iterable[float],
        backend: Optional[Backend] = None,
        auto_update: bool = True,
        absolute: bool = True,
        cal_parameter_name: Optional[str] = "drive_freq",
    ):
        """See :class:`.QubitSpectroscopy` for detailed documentation.

        Args:
            physical_qubits: List with the qubit on which to run spectroscopy.
            calibrations: If calibrations is given then running the experiment may update the values
                of the frequencies stored in calibrations.
            frequencies: The frequencies to scan in the experiment, in Hz.
            backend: Optional, the backend to run the experiment on.
            auto_update: If set to True, which is the default, then the experiment will
                automatically update the frequency in the calibrations.
            absolute: Boolean to specify if the frequencies are absolute or relative to the
                qubit frequency in the backend.
            cal_parameter_name: The name of the parameter to update in the calibrations.
                This defaults to `drive_freq`.

        Raises:
            QiskitError: If there are less than three frequency shifts.

        """
        super().__init__(
            calibrations,
            physical_qubits,
            frequencies,
            backend=backend,
            absolute=absolute,
            updater=Frequency,
            auto_update=auto_update,
            cal_parameter_name=cal_parameter_name,
        )

    def _attach_calibrations(self, circuit: QuantumCircuit):
        """QubitSpectroscopy already has the schedules attached in the program circuits."""
        pass


class RoughEFFrequencyCal(BaseCalibrationExperiment, EFSpectroscopy):
    r"""A calibration experiment that runs :class:`.QubitSpectroscopy` for the
    :math:`|1\rangle` <-> :math:`|2\rangle` transition.
    """

    __updater__ = Frequency

    # pylint: disable=super-init-not-called
    def __init__(
        self,
        physical_qubits: Sequence[int],
        calibrations: Calibrations,
        frequencies: Iterable[float],
        backend: Optional[Backend] = None,
        auto_update: bool = True,
        absolute: bool = True,
    ):
        """See :class:`.QubitSpectroscopy` for detailed documentation.

        Args:
            physical_qubits: List containing the qubit on which to run spectroscopy.
            calibrations: If calibrations is given then running the experiment may update the values
                of the frequencies stored in calibrations.
            frequencies: The frequencies to scan in the experiment, in Hz.
            backend: Optional, the backend to run the experiment on.
            auto_update: If set to True, which is the default, then the experiment will
                automatically update the frequency in the calibrations.
            absolute: Boolean to specify if the frequencies are absolute or relative to the
                qubit frequency in the backend.

        Raises:
            QiskitError: If there are less than three frequency shifts.

        """
        super().__init__(
            calibrations,
            physical_qubits,
            frequencies,
            backend,
            absolute,
            cal_parameter_name="f12",
            updater=Frequency,
            auto_update=auto_update,
        )

    def _attach_calibrations(self, circuit: QuantumCircuit):
        """Adds the calibrations to the transpiled circuits."""
        schedule = self._cals.get_schedule("x", self.physical_qubits)
        circuit.add_calibration("x", self.physical_qubits, schedule)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Quantum process tomography analysis
"""
from qiskit_experiments.framework import Options
from .basis import PauliMeasurementBasis, PauliPreparationBasis
from .tomography_analysis import TomographyAnalysis


class ProcessTomographyAnalysis(TomographyAnalysis):
    """Quantum state and process tomography experiment analysis.

    # section: overview

        Fitter Functions

        Built-in fitter functions may be selected using the following string
        labels, refer to the corresponding functions documentation for additional
        details on the fitters.

        * ``"linear_inversion"``:
          :func:`~qiskit_experiments.library.tomography.fitters.linear_inversion` (Default)
        * ``"cvxpy_linear_lstsq"``:
          :func:`~qiskit_experiments.library.tomography.fitters.cvxpy_linear_lstsq`
        * ``"cvxpy_gaussian_lstsq"``:
          :func:`~qiskit_experiments.library.tomography.fitters.cvxpy_gaussian_lstsq`

        PSD Rescaling

        For fitters that do not constrain the reconstructed state to be
        `positive-semidefinite` (PSD) we construct the maximum-likelihood
        nearest PSD state under the assumption of Gaussian measurement noise
        using the rescaling method in Reference [1]. For fitters that already
        support PSD constraints this option can be disabled by setting
        ``rescale_positive=False``.

    # section: note
        Fitters starting with ``"cvxpy_*"`` require the optional CVXPY Python
        package to be installed.

    # section: warning
        The API for tomography fitters is still under development so may change
        in future releases.

    # section: reference
        .. ref_arxiv:: 1 1106.5458

    """

    @classmethod
    def _default_options(cls) -> Options:
        """Default analysis options

        Analysis Options:
            measurement_basis (:class:`~qiskit_experiments.library.tomography.basis.MeasurementBasis`):
                The measurement
                :class:`~qiskit_experiments.library.tomography.basis.MeasurementBasis`
                to use for tomographic process reconstruction.
            preparation_basis (:class:`~qiskit_experiments.library.tomography.basis.PreparationBasis`):
                The preparation
                :class:`~qiskit_experiments.library.tomography.basis.PreparationBasis`
                to use for tomographic process reconstruction.
            fitter (str or Callable): The fitter function to use for reconstruction.
                rescale_psd (bool): If True rescale the fitted state to be positive-semidefinite
                (Default: True).
            fitter_options (dict): Any addition kwarg options to be supplied to the fitter
                function. For documentation of available kwargs refer to the fitter function
                documentation.
            rescale_positive (bool): If True rescale the state returned by the fitter
                to be positive-semidefinite. See the `PSD Rescaling` section for
                additional information (Default: True).
            rescale_trace (bool): If True rescale the state returned by the fitter
                have either trace 1 for :class:`~qiskit.quantum_info.DensityMatrix`,
                or trace dim for :class:`~qiskit.quantum_info.Choi` matrices (Default: True).
            measurement_qubits (Sequence[int]): Optional, the physical qubits with tomographic
                measurements. If not specified will be set to ``[0, ..., N-1]`` for N-qubit
                tomographic measurements.
            preparation_qubits (Sequence[int]): Optional, the physical qubits with tomographic
                preparations. If not specified will be set to ``[0, ..., N-1]`` for N-qubit
                tomographic preparations.
            fitter (str or Callable): The fitter function to use for reconstruction.
                This can be a string to select one of the built-in fitters, or a callable to
                supply a custom fitter function. See the `Fitter Functions` section for
                additional information.
            target (str or
                :class:`~qiskit.quantum_info.operators.channel.quantum_channel.QuantumChannel`
                or :class:`~qiskit.quantum_info.Operator`): Optional, Set a custom target quantum
                channel for computing the :func:`~qiskit.quantum_info.process_fidelity` of the
                fitted process against (Default: None).
            conditional_circuit_clbits (list[int]): Optional, the clbit indices in the
                source circuit to be conditioned on when reconstructing the channel.
                Enabling this will return a list of reconstructed channel components
                conditional on the values of these clbit values. The integer value of the
                conditioning clbits is stored in state analysis result extra field
                `"conditional_circuit_outcome"`.
            conditional_measurement_indices (list[int]): Optional, indices of tomography
                measurement qubits to used for conditional state reconstruction. Enabling
                this will return a list of reconstructed channel components conditioned on
                the remaining tomographic bases conditional on the basis index, and outcome
                value for these measurements. The conditional measurement basis index and
                integer value of the measurement outcome is stored in state analysis result
                extra fields `"conditional_measurement_index"` and
                `"conditional_measurement_outcome"` respectively.
            conditional_preparation_indices (list[int]): Optional, indices of tomography
                preparation qubits to used for conditional state reconstruction. Enabling
                this will return a list of reconstructed channel components conditioned on
                the remaining tomographic bases conditional on the basis index. The
                conditional preparation basis index is stored in state analysis result
                extra fields `"conditional_preparation_index"`.
        """
        options = super()._default_options()
        options.measurement_basis = PauliMeasurementBasis()
        options.preparation_basis = PauliPreparationBasis()
        return options



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Standard RB Experiment class.
"""
import functools
import logging
from collections import defaultdict
from numbers import Integral
from typing import Union, Iterable, Optional, List, Sequence, Dict, Any

import numpy as np
import rustworkx as rx
from numpy.random import Generator, default_rng
from numpy.random.bit_generator import BitGenerator, SeedSequence

from qiskit.circuit import CircuitInstruction, QuantumCircuit, Instruction, Barrier, Gate
from qiskit.exceptions import QiskitError
from qiskit.providers import BackendV2Converter
from qiskit.providers.backend import Backend, BackendV1, BackendV2
from qiskit.pulse.instruction_schedule_map import CalibrationPublisher
from qiskit.quantum_info import Clifford
from qiskit.quantum_info.random import random_clifford
from qiskit.transpiler import CouplingMap
from qiskit_experiments.framework import BaseExperiment, Options
from qiskit_experiments.framework.restless_mixin import RestlessMixin
from .clifford_utils import (
    CliffordUtils,
    DEFAULT_SYNTHESIS_METHOD,
    compose_1q,
    compose_2q,
    inverse_1q,
    inverse_2q,
    _clifford_1q_int_to_instruction,
    _clifford_2q_int_to_instruction,
    _clifford_to_instruction,
    _transpile_clifford_circuit,
)
from .rb_analysis import RBAnalysis

LOG = logging.getLogger(__name__)


SequenceElementType = Union[Clifford, Integral, QuantumCircuit]


class StandardRB(BaseExperiment, RestlessMixin):
    """An experiment to characterize the error rate of a gate set on a device.

    # section: overview

    Randomized Benchmarking (RB) is an efficient and robust method
    for estimating the average error rate of a set of quantum gate operations.
    See `Qiskit Textbook
    <https://github.com/Qiskit/textbook/blob/main/notebooks/quantum-hardware/randomized-benchmarking.ipynb>`_
    for an explanation on the RB method.

    A standard RB experiment generates sequences of random Cliffords
    such that the unitary computed by the sequences is the identity.
    After running the sequences on a backend, it calculates the probabilities to get back to
    the ground state, fits an exponentially decaying curve, and estimates
    the Error Per Clifford (EPC), as described in Refs. [1, 2].

    .. note::
        In 0.5.0, the default value of ``optimization_level`` in ``transpile_options`` changed
        from ``0`` to ``1`` for RB experiments. That may result in shorter RB circuits
        hence slower decay curves than before.

    # section: analysis_ref
        :class:`RBAnalysis`

    # section: manual
        :doc:`/manuals/verification/randomized_benchmarking`

    # section: reference
        .. ref_arxiv:: 1 1009.3639
        .. ref_arxiv:: 2 1109.6887
    """

    def __init__(
        self,
        physical_qubits: Sequence[int],
        lengths: Iterable[int],
        backend: Optional[Backend] = None,
        num_samples: int = 3,
        seed: Optional[Union[int, SeedSequence, BitGenerator, Generator]] = None,
        full_sampling: Optional[bool] = False,
    ):
        """Initialize a standard randomized benchmarking experiment.

        Args:
            physical_qubits: List of physical qubits for the experiment.
            lengths: A list of RB sequences lengths.
            backend: The backend to run the experiment on.
            num_samples: Number of samples to generate for each sequence length.
            seed: Optional, seed used to initialize ``numpy.random.default_rng``.
                  when generating circuits. The ``default_rng`` will be initialized
                  with this seed value every time :meth:`circuits` is called.
            full_sampling: If True all Cliffords are independently sampled for all lengths.
                           If False for sample of lengths longer sequences are constructed
                           by appending additional samples to shorter sequences.
                           The default is False.

        Raises:
            QiskitError: If any invalid argument is supplied.
        """
        # Initialize base experiment
        super().__init__(physical_qubits, analysis=RBAnalysis(), backend=backend)

        # Verify parameters
        if any(length <= 0 for length in lengths):
            raise QiskitError(
                f"The lengths list {lengths} should only contain " "positive elements."
            )
        if len(set(lengths)) != len(lengths):
            raise QiskitError(
                f"The lengths list {lengths} should not contain " "duplicate elements."
            )
        if num_samples <= 0:
            raise QiskitError(f"The number of samples {num_samples} should " "be positive.")

        # Set configurable options
        self.set_experiment_options(
            lengths=sorted(lengths), num_samples=num_samples, seed=seed, full_sampling=full_sampling
        )
        self.analysis.set_options(outcome="0" * self.num_qubits)

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            lengths (List[int]): A list of RB sequences lengths.
            num_samples (int): Number of samples to generate for each sequence length.
            seed (None or int or SeedSequence or BitGenerator or Generator): A seed
                used to initialize ``numpy.random.default_rng`` when generating circuits.
                The ``default_rng`` will be initialized with this seed value every time
                :meth:`circuits` is called.
            full_sampling (bool): If True all Cliffords are independently sampled for
                all lengths. If False for sample of lengths longer sequences are constructed
                by appending additional Clifford samples to shorter sequences.
            clifford_synthesis_method (str): The name of the Clifford synthesis plugin to use
                for building circuits of RB sequences.
        """
        options = super()._default_experiment_options()
        options.update_options(
            lengths=None,
            num_samples=None,
            seed=None,
            full_sampling=None,
            clifford_synthesis_method=DEFAULT_SYNTHESIS_METHOD,
        )

        return options

    @classmethod
    def _default_transpile_options(cls) -> Options:
        """Default transpiler options for transpiling RB circuits."""
        return Options(optimization_level=1)

    def _set_backend(self, backend: Backend):
        """Set the backend V2 for RB experiments since RB experiments only support BackendV2
        except for simulators. If BackendV1 is provided, it is converted to V2 and stored.
        """
        if isinstance(backend, BackendV1) and "simulator" not in backend.name():
            super()._set_backend(BackendV2Converter(backend, add_delay=True))
        else:
            super()._set_backend(backend)

    def circuits(self) -> List[QuantumCircuit]:
        """Return a list of RB circuits.

        Returns:
            A list of :class:`QuantumCircuit`.
        """
        # Sample random Clifford sequences
        sequences = self._sample_sequences()
        # Convert each sequence into circuit and append the inverse to the end.
        circuits = self._sequences_to_circuits(sequences)
        # Add metadata for each circuit
        for circ, seq in zip(circuits, sequences):
            circ.metadata = {
                "xval": len(seq),
                "group": "Clifford",
            }
        return circuits

    def _sample_sequences(self) -> List[Sequence[SequenceElementType]]:
        """Sample RB sequences

        Returns:
            A list of RB sequences.
        """
        rng = default_rng(seed=self.experiment_options.seed)
        sequences = []
        if self.experiment_options.full_sampling:
            for _ in range(self.experiment_options.num_samples):
                for length in self.experiment_options.lengths:
                    sequences.append(self.__sample_sequence(length, rng))
        else:
            for _ in range(self.experiment_options.num_samples):
                longest_seq = self.__sample_sequence(max(self.experiment_options.lengths), rng)
                for length in self.experiment_options.lengths:
                    sequences.append(longest_seq[:length])

        return sequences

    def _get_synthesis_options(self) -> Dict[str, Optional[Any]]:
        """Get options for Clifford synthesis from the backend information as a dictionary.

        The options include:
        - "basis_gates": Sorted basis gate names.
            Return None if no basis gates are supplied via ``backend`` or ``transpile_options``.
        - "coupling_tuple": Reduced coupling map in the form of tuple of edges in the coupling graph.
            Return None if no coupling map are supplied via ``backend`` or ``transpile_options``.

        Returns:
            Synthesis options as a dictionary.
        """
        basis_gates = self.transpile_options.get("basis_gates", [])
        coupling_map = self.transpile_options.get("coupling_map", None)
        if coupling_map:
            coupling_map = coupling_map.reduce(self.physical_qubits)
        if not (basis_gates and coupling_map) and self.backend:
            if isinstance(self.backend, BackendV2) and "simulator" in self.backend.name:
                basis_gates = basis_gates if basis_gates else self.backend.target.operation_names
                coupling_map = coupling_map if coupling_map else None
            elif isinstance(self.backend, BackendV2):
                gate_ops = [op for op in self.backend.target.operations if isinstance(op, Gate)]
                backend_basis_gates = [op.name for op in gate_ops if op.num_qubits != 2]
                backend_cmap = None
                for op in gate_ops:
                    if op.num_qubits != 2:
                        continue
                    cmap = self.backend.target.build_coupling_map(op.name)
                    if cmap is None:
                        backend_basis_gates.append(op.name)
                    else:
                        reduced = cmap.reduce(self.physical_qubits)
                        if rx.is_weakly_connected(reduced.graph):
                            backend_basis_gates.append(op.name)
                            backend_cmap = reduced
                            # take the first non-global 2q gate if backend has multiple 2q gates
                            break
                basis_gates = basis_gates if basis_gates else backend_basis_gates
                coupling_map = coupling_map if coupling_map else backend_cmap
            elif isinstance(self.backend, BackendV1):
                backend_basis_gates = self.backend.configuration().basis_gates
                backend_cmap = self.backend.configuration().coupling_map
                if backend_cmap:
                    backend_cmap = CouplingMap(backend_cmap).reduce(self.physical_qubits)
                basis_gates = basis_gates if basis_gates else backend_basis_gates
                coupling_map = coupling_map if coupling_map else backend_cmap

        return {
            "basis_gates": tuple(sorted(basis_gates)) if basis_gates else None,
            "coupling_tuple": tuple(sorted(coupling_map.get_edges())) if coupling_map else None,
            "synthesis_method": self.experiment_options["clifford_synthesis_method"],
        }

    def _sequences_to_circuits(
        self, sequences: List[Sequence[SequenceElementType]]
    ) -> List[QuantumCircuit]:
        """Convert an RB sequence into circuit and append the inverse to the end.

        Returns:
            A list of RB circuits.
        """
        synthesis_opts = self._get_synthesis_options()
        # Circuit generation
        circuits = []
        for i, seq in enumerate(sequences):
            if (
                self.experiment_options.full_sampling
                or i % len(self.experiment_options.lengths) == 0
            ):
                prev_elem, prev_seq = self.__identity_clifford(), []

            circ = QuantumCircuit(self.num_qubits)
            for elem in seq:
                circ.append(self._to_instruction(elem, synthesis_opts), circ.qubits)
                circ._append(CircuitInstruction(Barrier(self.num_qubits), circ.qubits))

            # Compute inverse, compute only the difference from the previous shorter sequence
            prev_elem = self.__compose_clifford_seq(prev_elem, seq[len(prev_seq) :])
            prev_seq = seq
            inv = self.__adjoint_clifford(prev_elem)

            circ.append(self._to_instruction(inv, synthesis_opts), circ.qubits)
            circ.measure_all()  # includes insertion of the barrier before measurement
            circuits.append(circ)
        return circuits

    def __sample_sequence(self, length: int, rng: Generator) -> Sequence[SequenceElementType]:
        # Sample an RB sequence with the given length.
        # Return integer instead of Clifford object for 1 or 2 qubits case for speed
        if self.num_qubits == 1:
            return rng.integers(CliffordUtils.NUM_CLIFFORD_1_QUBIT, size=length)
        if self.num_qubits == 2:
            return rng.integers(CliffordUtils.NUM_CLIFFORD_2_QUBIT, size=length)
        # Return Clifford object for 3 or more qubits case
        return [random_clifford(self.num_qubits, rng) for _ in range(length)]

    def _to_instruction(
        self,
        elem: SequenceElementType,
        synthesis_options: Dict[str, Optional[Any]],
    ) -> Instruction:
        """Return the instruction of a Clifford element.

        The resulting instruction contains a circuit definition with ``basis_gates`` and
        it complies with ``coupling_tuple``, which is specified in ``synthesis_options``.

        Args:
            elem: a Clifford element to be converted
            synthesis_options: options for synthesizing the Clifford element

        Returns:
            Converted instruction
        """
        # Switching for speed up
        if isinstance(elem, Integral):
            if self.num_qubits == 1:
                return _clifford_1q_int_to_instruction(
                    elem,
                    basis_gates=synthesis_options["basis_gates"],
                    synthesis_method=synthesis_options["synthesis_method"],
                )
            if self.num_qubits == 2:
                return _clifford_2q_int_to_instruction(elem, **synthesis_options)

        return _clifford_to_instruction(elem, **synthesis_options)

    def __identity_clifford(self) -> SequenceElementType:
        if self.num_qubits <= 2:
            return 0
        return Clifford(np.eye(2 * self.num_qubits))

    def __compose_clifford_seq(
        self, base_elem: SequenceElementType, elements: Sequence[SequenceElementType]
    ) -> SequenceElementType:
        if self.num_qubits <= 2:
            return functools.reduce(
                compose_1q if self.num_qubits == 1 else compose_2q, elements, base_elem
            )
        # 3 or more qubits
        res = base_elem
        for elem in elements:
            res = res.compose(elem)
        return res

    def __adjoint_clifford(self, op: SequenceElementType) -> SequenceElementType:
        if self.num_qubits == 1:
            return inverse_1q(op)
        if self.num_qubits == 2:
            return inverse_2q(op)
        if isinstance(op, QuantumCircuit):
            return Clifford.from_circuit(op).adjoint()
        return op.adjoint()

    def _transpiled_circuits(self) -> List[QuantumCircuit]:
        """Return a list of experiment circuits, transpiled."""
        has_custom_transpile_option = (
            not set(vars(self.transpile_options)).issubset(
                {"basis_gates", "coupling_map", "optimization_level"}
            )
            or self.transpile_options.get("optimization_level", 1) != 1
        )
        if has_custom_transpile_option:
            transpiled = super()._transpiled_circuits()
        else:
            transpiled = [
                _transpile_clifford_circuit(circ, physical_qubits=self.physical_qubits)
                for circ in self.circuits()
            ]
            # Set custom calibrations provided in backend (excluding simulators)
            if isinstance(self.backend, BackendV2) and "simulator" not in self.backend.name:
                qargs_patterns = [self.physical_qubits]  # for 1q or 3q+ case
                if self.num_qubits == 2:
                    qargs_patterns = [
                        (self.physical_qubits[0],),
                        (self.physical_qubits[1],),
                        self.physical_qubits,
                        (self.physical_qubits[1], self.physical_qubits[0]),
                    ]

                qargs_supported = self.backend.target.qargs
                instructions = []  # (op_name, qargs) for each element where qargs means qubit tuple
                for qargs in qargs_patterns:
                    if qargs in qargs_supported:
                        for op_name in self.backend.target.operation_names_for_qargs(qargs):
                            instructions.append((op_name, qargs))

                common_calibrations = defaultdict(dict)
                for op_name, qargs in instructions:
                    inst_prop = self.backend.target[op_name].get(qargs, None)
                    if inst_prop is None:
                        continue
                    schedule = inst_prop.calibration
                    if schedule is None:
                        continue
                    publisher = schedule.metadata.get("publisher", CalibrationPublisher.QISKIT)
                    if publisher != CalibrationPublisher.BACKEND_PROVIDER:
                        common_calibrations[op_name][(qargs, tuple())] = schedule

                for circ in transpiled:
                    # This logic is inefficient in terms of payload size and backend compilation
                    # because this binds every custom pulse to a circuit regardless of
                    # its existence. It works but redundant calibration must be removed -- NK.
                    circ.calibrations = common_calibrations

        if self.analysis.options.get("gate_error_ratio", None) is None:
            # Gate errors are not computed, then counting ops is not necessary.
            return transpiled

        # Compute average basis gate numbers per Clifford operation
        # This is probably main source of performance regression.
        # This should be integrated into transpile pass in future.
        qubit_indices = {bit: index for index, bit in enumerate(transpiled[0].qubits)}
        for circ in transpiled:
            count_ops_result = defaultdict(int)
            # This is physical circuits, i.e. qargs is physical index
            for cdata in circ.data:
                inst = cdata.operation
                qargs = cdata.qubits
                if inst.name in ("measure", "reset", "delay", "barrier", "snapshot"):
                    continue
                qinds = [qubit_indices[q] for q in qargs]
                if not set(self.physical_qubits).issuperset(qinds):
                    continue
                # Not aware of multi-qubit gate direction
                formatted_key = tuple(sorted(qinds)), inst.name
                count_ops_result[formatted_key] += 1
            circ.metadata["count_ops"] = tuple(count_ops_result.items())

        return transpiled

    def _metadata(self):
        metadata = super()._metadata()
        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)

        return metadata
